import copy
import time
from collections import defaultdict, Counter
from typing import Dict, List, Tuple, Optional, Union

import scipy as sc
import numpy as np
from tqdm import tqdm

from qrem.functions_qrem import povmtools
from qrem.functions_qrem import ancillary_functions as anf
from qrem.noise_simulation.CN import functions_sampling as fus

import re

from qrem.common import convert
from qrem.common import utils


__parities__dict = {1: {('0',): 1,
                        ('1',): -1},

                    2: {('0', '0'): 1,
                        ('1', '1'): 1,
                        ('0', '1'): -1,
                        ('1', '0'): -1,

                        }

                    }

_parities__dict2 = {'0': 1,
                    '1': -1,
                    '00': 1,
                    '11': 1,
                    '01': -1,
                    '10': -1,

                    }

#different normalisation
_parities__dict3 = {'0': 0,
                    '1': 1,
                    '00': 0,
                    '11': 0,
                    '01': 1,
                    '10': 1,

                    }



#ORGANIZE - this function computes energy of some local hamiltonians form marginals, this is used in benchmark functions probably (MO)
def estimate_energy_from_marginals(weights_dictionary: Dict[Tuple[int], float],
                                   marginals_dictionary: Dict[Tuple[int], np.ndarray],
                                   additional_multipliers=None):
    """
    Compute energy of Hamiltonian from dictionary of marginal distributions.

    :param weights_dictionary:
    :param marginals_dictionary:
    :return:
    """

    energy = 0
    for key_local_term in weights_dictionary.keys():
        weight = weights_dictionary[key_local_term]
        marginal = marginals_dictionary[key_local_term]

        # print(marginal,len(marginal))
        qubits_number = int(np.log2(len(marginal)))

        local_term_energy = 0
        for result_index in range(len(marginal)):
            bitstring = list(convert.integer_to_bitstring(result_index, qubits_number))
            bit_true = [int(x) for x in bitstring]
            parity = (-1) ** (np.count_nonzero(bit_true))
            #
            # bitstring_getitem = convert.integer_to_bitstring(result_index, qubits_number).__getitem__
            # parity = __get_part_bitstring_parity_special(bitstring_getitem=bitstring_getitem,
            #                                              qubit_indices=range(qubits_number))
            #
            # print(bit_true,parity, weight, marginal[result_index])

            local_term_energy += weight * marginal[result_index] * parity
            # print(local_term_energy)
        if additional_multipliers is not None:
            local_term_energy *= additional_multipliers[key_local_term]

        energy += local_term_energy

    if isinstance(energy, list) or isinstance(energy, np.ndarray):
        return energy[0]
    else:
        return energy

def __get_part_bitstring_parity_special(bitstring_getitem,
                                        qubit_indices):
    """Used only in get_energy_from_bitstring_diagonal_local()"""
    return (-1) ** list(map(bitstring_getitem, qubit_indices)).count('1')

"""
def __get_part_bitstring_parity_special2(bitstring_getitem,
                                         qubit_indices):
    return __parities__dict[len(qubit_indices)][tuple(map(bitstring_getitem,
                                                    qubit_indices))]

def __get_part_bitstring_parity_special3(bitstring,
                                         qubit_indices):
    return __parities__dict[tuple([bitstring[qi] for qi in qubit_indices])]
"""

def __get_part_bitstring_parity_special4(bitstring,
                                         qubit_indices):
    """Used only in estimate_energy_from_counts_dictionary_alternative()"""
    return _parities__dict2[''.join(([bitstring[qi] for qi in qubit_indices]))]


#(PP) Hamiltonian w stylu Isinga
def get_energy_from_bitstring_diagonal_local(bitstring: str,
                                             weights_dictionary: Dict[Tuple[int], float],
                                             additional_multipliers=None
                                             ):
    """Calculates the energy corresponding to a state (encoded in bistring) and a Hamiltonian
    (encoded in weights and additional_multipliers)"""
    if isinstance(bitstring,list):
        bitstring = ''.join(bitstring)

    bitstring_getitem = bitstring.__getitem__

    parities = {qubit_indices: __get_part_bitstring_parity_special(bitstring_getitem,
                                                                   qubit_indices) for qubit_indices in
                weights_dictionary.keys()}

    energy = 0
    for qubit_indices, hamiltonian_coefficient in weights_dictionary.items():
        # marginal_bitstring = [int(bitstring[q]) for q in qubit_indices]
        # parity = (-1) ** (np.count_nonzero(marginal_bitstring))
        local_energy = parities[qubit_indices] * hamiltonian_coefficient
        if additional_multipliers is not None:
            local_energy *= additional_multipliers[qubit_indices]
        energy += local_energy

    return energy

#(PP) Hamiltonian w stylu Isinga
"""
def get_energy_from_bitstring_diagonal_global(bitstring: str,
                                             weights_dictionary: Dict[Tuple[int], float],
                                             additional_multipliers=None
                                             ):

    energy = 0
    for qubit_indices, hamiltonian_coefficient in weights_dictionary.items():
        marginal_bitstring = [int(bitstring[q]) for q in qubit_indices]
        parity = (-1) ** (np.count_nonzero(marginal_bitstring))
        local_energy = parity * hamiltonian_coefficient
        if additional_multipliers is not None:
            local_energy *= additional_multipliers[qubit_indices]
        energy += local_energy

    return energy
"""

def get_noisy_energy_product_noise_model(input_state,
                                         noise_matrices_dictionary,
                                         weights_dictionary_tuples,
                                         needed_pairs=None):
    pairs_marginals_noisy = fus.calculate_pairs_marginals_from_tensor_model_fixed_input(
        input_state=input_state,
        noise_matrices_dictionary=noise_matrices_dictionary,
        needed_pairs=needed_pairs,
        get_also_1q_marginals=True)

    energy_modeled_exact = estimate_energy_from_marginals(
        weights_dictionary=weights_dictionary_tuples,
        marginals_dictionary=pairs_marginals_noisy)

    return energy_modeled_exact

def estimate_energy_from_counts_dictionary(counts_dictionary: Dict[str, int],
                                           weights_dictionary: Dict[Tuple[int], float],
                                           additional_multipliers=None,
                                           normalize=True):
    """
    Calculates the energy averaged over results, used by the class EnergyEstimatorBase

    :param counts_dictionary:
    :type counts_dictionary:
    :param weights_dictionary:
    :type weights_dictionary:
    :return:
    :rtype:
    """

    energy = 0
    for bitstring, shots in counts_dictionary.items():
        energy += get_energy_from_bitstring_diagonal_local(bitstring=bitstring,
                                                           weights_dictionary=weights_dictionary,
                                                           additional_multipliers=additional_multipliers) * shots

    if normalize:
        normalization = sum(list(counts_dictionary.values()))
        energy /= normalization

    return energy

def estimate_energy_from_counts_dictionary_alternative(counts_dictionary: Dict[str, int],
                                                       weights_dictionary: Dict[Tuple[int], float],
                                                       additional_multipliers=None,
                                                       normalize: Optional[Union[bool, float]] = True,
                                                       print_progress_bar=False):
    """
    Alternative to estimate_energy_from_counts_dictionary()
    used ONLY in noise_mitigation/expected_values/CRS/testing_mitigation_clusters_SCR_amd.py

    :param counts_dictionary:
    :type counts_dictionary:
    :param weights_dictionary:
    :type weights_dictionary:
    :return:
    :rtype:
    """
    t0 = time.time()
    tests = True
    if tests:
        if additional_multipliers is not None:
            energy = 0
            for qubit_indices in tqdm(weights_dictionary.keys()):
                local_energy=0
                for bitstring, shots in counts_dictionary.items():
                    local_energy+= _parities__dict2[''.join([bitstring[qi] for qi in qubit_indices])] * shots

                energy+=local_energy*additional_multipliers[qubit_indices]*weights_dictionary[qubit_indices]
        else:
            energy = 0
            for qubit_indices in tqdm(weights_dictionary.keys(),disable=True):
                local_energy=0
                for bitstring, shots in counts_dictionary.items():
                    local_energy+= _parities__dict2[''.join([bitstring[qi] for qi in qubit_indices])] * shots

                energy += local_energy * weights_dictionary[
                    qubit_indices]

    else:

        parities = {}
        for bitstring in counts_dictionary.keys():
            parities[bitstring] = {
            qubit_indices: __get_part_bitstring_parity_special4(
                bitstring=bitstring,
                qubit_indices=qubit_indices) for
            qubit_indices in weights_dictionary.keys()}

        t1=time.time()
            # for subset in

        energy = 0
        if additional_multipliers is not None:
            for qubit_indices, hamiltonian_coefficient in tqdm(weights_dictionary.items(),
                                                               disable=not print_progress_bar):
                local_energy = 0
                for bitstring, shots in counts_dictionary.items():
                    local_energy += shots * parities[bitstring][qubit_indices]

                # local_energy = parities[qubit_indices][1]-parities[qubit_indices][-1]



                local_energy *= additional_multipliers[qubit_indices]
                local_energy *= hamiltonian_coefficient

                energy += local_energy
        else:
            for qubit_indices, hamiltonian_coefficient in tqdm(weights_dictionary.items(),
                                                               disable=not print_progress_bar):
                local_energy = 0
                for bitstring, shots in counts_dictionary.items():
                    local_energy += shots * parities[bitstring][qubit_indices]
                local_energy *= hamiltonian_coefficient

                energy += local_energy

        t2 = time.time()


    t1=time.time()
    if isinstance(normalize, float):
        energy /= normalize
    elif isinstance(normalize, bool):
        if normalize:
            normalization = sum(list(counts_dictionary.values()))
            energy /= normalization

    t2=time.time()

    # print(t1-t0)
    # print('hej')
    return energy


def merge_multiple_counts_dictionaries(counts_dictionaries_list: List[Dict[str, int]]) -> Dict[
    str, int]:
    """
    Merge multiple counts dictionaries.
    This is useful when you have results of multiple implementations of the same experiment.

    :param counts_dictionaries_list: list of results of counts dictionaries of the form:
                                    {'bitstring":number of measurements}
    :return:
    """

    # first dictionary will be template to which we will add counts
    merged_counts = copy.deepcopy(counts_dictionaries_list[0])

    # we go through all other dictionaries
    for counts_dictionary in counts_dictionaries_list[1:]:
        for bitstring, ticks in counts_dictionary.items():
            if bitstring in merged_counts.keys():
                merged_counts[bitstring] += ticks
            else:
                merged_counts[bitstring] = ticks

    return merged_counts

def reverse_bitstrings_in_counts_dictionary(counts_dictionary):
    """Used only by convert_counts_overlapping_tomography()"""
    counts_reversed = {}

    for bitstring, ticks in counts_dictionary.items():
        counts_reversed[bitstring[::-1]] = ticks
    return counts_reversed

def convert_counts_overlapping_tomography(counts_dictionary: Dict[str, Dict[str, int]],
                                          experiment_name: str,
                                          reverse_bitstrings=True,
                                          old_send_procedure=True):

    # TODO FBM: add support for Quantum Detector Tomography
    # MOcomm - this looks
    """
    This function converts unprocessed dictionary of experiment results, where multiple keys can describe identical
    circuits (eg. "DDOT-010no0", "DDOT-010no1", see description of string_cutter below), to a dictionary where a key
    uniquely corresponds to a circuit and value is the combined counts from all such circuits in the unprocessed
    dictionary. In resulting dictionary outcome bit strings and circuit names are in big-endian order.

    param counts_dictionary: dictionary where the key is circuit name (e.g. "DDOT-010no3", described in inner
                              string_cutter function below) and value is dictionary of counts, where the key is a string
                              denoting classical outcome and the value the number of its occurrences in given experiment.
    param experiment_name: string denoting the type of experiment whose results counts_dictionary contains, e.g. 'QDOT';
                            the valid names are specified in SeparableCircuitsCreator.__valid_experiments_names__
    param reverse_bitstrings: bool; if False, the bitstring denoting classical outcome in counts_dictionary will be
                               interpreted as big-endian (qubit 0 on the left); if True it will be interpreted as
                               little-endian (qubit 0 on the right) and will be reversed to conform to QREM convention.

    :return: big_counts_dictionary: dictionary where key is string describing circuit (e.g. '010' means gates:
                                    iden, X, iden on qubits 0, 1, 2 respectively) and value is
                                    dictionary where key is bitstring describing outcome (e.g. '100') and value is
                                    number of occurences of that outcome in the experiment; these strings are big-endian.
    """
    def string_cutter(circuit_name: str):
        """
        This inner function cuts the name of the circuit to the format that will later be used by
        tomography data analyzers.


        param circuit_name:
        It assumes the following convention:

        circuit_name = "experiment name" + "-" + "circuit label"+
        "no"+ "integer identifier for multiple implementations of the same circuit"

        for example the circuit can have name:
        "DDOT-010no3"

        which means that this experiment is Diagonal Detector Overlapping Tomography (DDOT),
        the circuit implements state "010" (i.e., gates iden, X, iden on qubits 0,1,2 - big-endian order), and
        in the whole circuits sets this is the 4th (we start counting from 0) circuit that implements
        that particular state.

        :return: big_counts_dictionary
        """
        from qrem.noise_characterization.tomography_design.overlapping import SeparableCircuitsCreator

        if experiment_name.lower() not in SeparableCircuitsCreator.__valid_experiments_names__:
            raise ValueError(f"ONLY the following experiments are supported:\n{SeparableCircuitsCreator.__valid_experiments_names__}")

        experiment_string_len = len(list(experiment_name))
        full_name_now = circuit_name[experiment_string_len + 1:]
        new_string = ''
        for symbol_index in range(len(full_name_now)):
            if full_name_now[symbol_index] + full_name_now[symbol_index + 1] == 'no':
                break
            new_string += full_name_now[symbol_index]
        return new_string

    big_counts_dictionary = {}

    for circuit_name, counts_dict_now in counts_dictionary.items():
        
        #the line below was used with data obtained with the old sending routine  
        
        if old_send_procedure:
            proper_name_now = string_cutter(circuit_name)
        else: 
            m=re.search(r'\d+', circuit_name)
            proper_name_now=m.group()
        if reverse_bitstrings:
            counts_dict_now = reverse_bitstrings_in_counts_dictionary(counts_dict_now)

        if proper_name_now not in big_counts_dictionary.keys():
            big_counts_dictionary[proper_name_now] = counts_dict_now
        else:
            big_counts_dictionary[proper_name_now] = merge_multiple_counts_dictionaries(
                [big_counts_dictionary[proper_name_now], counts_dict_now])

    return big_counts_dictionary


def convert_subset_counts_dictionary_to_probability_distribution(counts_dictionary):
    """Makes a probability distribution from counts dictionary,
    IMPORTANT: fills in with zeros for results absent in counts_dictionary
    USED A LOT by functions from cn\mitigation.py and functions_qrem\new_mitigation_routines.py"""
    key = next(iter(counts_dictionary))
    subset_counts_dictionary = counts_dictionary[key]
    subset = next(iter(subset_counts_dictionary))
    number_of_qubits = len(subset) 
    probability_distribution = np.zeros((2 ** number_of_qubits), dtype=float)
    
    normalization_shots = sum(list(subset_counts_dictionary[subset]))
    probability_distribution= counts_dictionary[key][subset] / normalization_shots



    return  probability_distribution


def get_marginal_from_probability_distribution(
        global_probability_distribution: np.ndarray,
        bits_of_interest: List[int],
        register_names: Optional[List[str]] = None) -> np.ndarray:
    """Return marginal distribution from vector of global distribution
    :param global_probability_distribution: distribution on all bits
    :param bits_of_interest: bits we are interested in (so we average over other bits)
                            Assuming that qubits are labeled
                            from 0 to log2(len(global_probability_distribution))
    :param register_names: bitstrings register, default is
                           '00...00', '000...01', '000...10', ..., etc.

    :return: marginal_distribution : marginal probability distribution

    NOTE: we identify bits with qubits in the variables bitstring_names

    #TODO FBM: do some speed tests on some details of those solutions
    """

    if len(bits_of_interest) == 0:
        print('0 length bits list')
        return global_probability_distribution

    try:
        if isinstance(global_probability_distribution[0][0], complex) or isinstance(
                global_probability_distribution[0][0], np.complex128):
            global_probability_distribution = global_probability_distribution.real
    except(IndexError):
        if isinstance(global_probability_distribution[0], complex) or isinstance(
                global_probability_distribution[0], np.complex128):
            global_probability_distribution = global_probability_distribution.real

    global_dimension = len(global_probability_distribution)
    global_number_of_qubits = int(np.log2(global_dimension))
    all_qubits = list(range(global_number_of_qubits))
    bits_to_average_over = list(set(all_qubits).difference(set(bits_of_interest)))

    number_of_bits_in_marginal = global_number_of_qubits - len(bits_to_average_over)
    dimension_of_marginal = 2 ** number_of_bits_in_marginal

    if register_names is None:
        bitstring_names = anf.get_classical_register_bitstrings(list(range(global_number_of_qubits)),
                                                                global_number_of_qubits)
    else:
        bitstring_names = register_names

    marginal_distribution = np.zeros((dimension_of_marginal, 1), dtype=float)
    for j in range(global_dimension):
        # this is slightly faster than going through "for bitstring_global in bitstring_names
        # and then converting bitstring_global to integer
        # and also faster than creating the global bitstring in situ
        bitstring_global = bitstring_names[j]

        bitstring_local = ''.join(
            [bitstring_global[qubit_index] for qubit_index in bits_of_interest])

        marginal_distribution[int(bitstring_local, 2)] += global_probability_distribution[j]

    return marginal_distribution


def _sort_clusters_division(clusters_division: List[Tuple[int]]):
    sorted_inside = [tuple(sorted(x)) for x in clusters_division]

    return tuple(sorted(sorted_inside, key=lambda x: x[0]))


def get_noise_matrices_from_POVMs_dictionary(POVMs_dictionary):
    """Takes POVM dictionary and returns noise matrices,
    IMPORTANT: used by characterization\characterization_routine.py"""
    noise_matrices_dictionary = {}

    for qubits_subset, POVM in tqdm(POVMs_dictionary.items()):
        noise_matrices_dictionary[qubits_subset] = povmtools.get_noise_matrix_from_povm(povm=POVM)

    return noise_matrices_dictionary

def get_multiple_mitigation_strategies_clusters_for_pairs_of_qubits(pairs_of_qubits,
                                                           clusters_sets,
                                                     dictionary_results,
                                                     #marginals_dictionary,
                                                     noise_matrices_dictionary,
                                                    show_progress_bar = True
                                                     ):
    from qrem.noise_mitigation.probability_distributions.CorrectionDataGenerator import \
        CorrectionDataGenerator

    number_of_qubits = sum([len(x) for x in list(clusters_sets.keys())[0]])

    joint_correction_matrices_dictionary = {}
    correction_indices = {}
    mitigation_data ={}
    test_check={}
    for clusters_list, model_parameters in tqdm(clusters_sets.items(), disable = not show_progress_bar):
        neighborhoods = {cluster: [] for cluster in clusters_list}

        correction_data_generator = CorrectionDataGenerator(results_dictionary_ddt=dictionary_results,
                                                            #marginals_dictionary=marginals_dictionary,
                                                            number_of_qubits=number_of_qubits,
                                                            clusters_list=clusters_list,
                                                            neighborhoods=neighborhoods,
                                                            noise_matrices_dictionary=
                                                            noise_matrices_dictionary,
                                                            #TJT: his is probably not a good idea
                                                            #correction_matrices_dictionary=joint_correction_matrices_dictionary# )
                                                            correction_matrices_dictionary={})

        correction_data_generator.get_pairs_correction_data(pairs_list=pairs_of_qubits,
                                                            show_progress_bar=False)
        
        
        cluster_check={}
        for subset in  correction_data_generator._correction_matrices.keys():
            cluster_check[subset]=(correction_data_generator._noise_matrices_dictionary[subset]['averaged'].dot(correction_data_generator._correction_matrices[subset]))
        test_check[clusters_list] = cluster_check
            
                
         

        #JT: Here we use a private property _correction_matrices, _correction_indices of CorrectionDataGenerator class, thsi needs to be fixed  
        c_list = [correction_data_generator._correction_indices, correction_data_generator._correction_matrices]
        c_tuple=tuple(c_list)
        mitigation_data[clusters_list] = c_tuple
        
        
        joint_correction_matrices_dictionary = {**joint_correction_matrices_dictionary,
                                                **correction_data_generator._correction_matrices}

        correction_indices_now = correction_data_generator._correction_indices

        correction_indices[clusters_list] = correction_indices_now


    return joint_correction_matrices_dictionary, correction_indices, mitigation_data


# """
# def get_CTMP_rates_from_results(results_dictionary_ddot: Dict[str, Dict[str, int]],
#                                 number_of_qubits: int):
#     single_qubits = list(range(number_of_qubits))
#     pairs = [(i, j) for i in range(number_of_qubits) for j in range(i + 1, number_of_qubits)]

#     local_noise_matrices_CTMP = {pair: np.zeros((4, 4), dtype=float) for pair in pairs}

#     for pair in pairs:
#         pair_complement = list(set(single_qubits).difference(set(pair)))

#         for global_input_state, results_dictionary in results_dictionary_ddot.items():
#             marginal_input_state = ''.join([global_input_state[x] for x in pair])
#             input_state_complement = [global_input_state[x] for x in pair_complement]

#             for global_output_state, ticks in results_dictionary.items():
#                 marginal_output_state = ''.join([global_output_state[x] for x in pair])
#                 output_state_complement = [global_output_state[x] for x in pair_complement]

#                 if output_state_complement == input_state_complement:
#                     # check if this is their convention!
#                     local_noise_matrices_CTMP[pair][int(marginal_output_state, 2),
#                                                     int(marginal_input_state, 2)] += ticks

#     # normalize to stochastic matrices
#     for pair in pairs:
#         for k in range(4):
#             local_noise_matrices_CTMP[pair][:, k] /= sum(local_noise_matrices_CTMP[pair][:, k])

#     # Get G matrices
#     G_matrices = {pair: sc.linalg.logm(local_noise_matrices_CTMP[pair]) for pair in pairs}

#     # ancillary function
#     def _chop_negatives(M):
#         (m, n) = M.shape

#         chopped_M = copy.deepcopy(M)
#         for i in range(m):
#             for j in range(n):
#                 if i != j and M[i, j] < 0:
#                     chopped_M[i, j] = 0

#         return chopped_M

#     # Get G' matrices
#     G_prime_matrices = {pair: _chop_negatives(G_matrices[pair]) for pair in pairs}

#     rates_dictionary_1q = {f"q{q}": {'0': 0, '1': 0} for q in single_qubits}
#     for qj in single_qubits:
#         r0, r1 = 0, 0

#         for q_other in single_qubits:
#             if q_other != qj:
#                 G_prime_matrix_now = G_prime_matrices[tuple(sorted([qj, q_other]))]

#                 r0 += G_prime_matrix_now[2, 0] + G_prime_matrix_now[3, 1] \
#                     # +G_prime_matrix_now[1,0]+G_prime_matrix_now[3,2]

#                 r1 += G_prime_matrix_now[0, 2] + G_prime_matrix_now[1, 3] \
#                     # +G_prime_matrix_now[0, 1] + G_prime_matrix_now[2, 3]

#         r0 /= 2 * (number_of_qubits - 1)
#         r1 /= 2 * (number_of_qubits - 1)

#         rates_dictionary_1q[f"q{qj}"]['0'] = r0
#         rates_dictionary_1q[f"q{qj}"]['0'] = r1

#     rates_dictionary_2q = {f"q{pair[0]}q{pair[1]}": {'00': 0,
#                                                      '01': 0,
#                                                      '10': 0,
#                                                      '11': 0} for pair in pairs}

#     for pair in pairs:
#         G_prime_matrix_now = G_prime_matrices[pair]
#         rates_dictionary_2q[f"q{pair[0]}q{pair[1]}"]['00'] = G_prime_matrix_now[3, 0]
#         rates_dictionary_2q[f"q{pair[0]}q{pair[1]}"]['01'] = G_prime_matrix_now[2, 1]
#         rates_dictionary_2q[f"q{pair[0]}q{pair[1]}"]['10'] = G_prime_matrix_now[1, 2]
#         rates_dictionary_2q[f"q{pair[0]}q{pair[1]}"]['11'] = G_prime_matrix_now[0, 3]

#     rates_dictionary = {**rates_dictionary_1q,
#                         **rates_dictionary_2q}

#     return rates_dictionary


# def convert_counts_dictionary_to_probability_distribution(counts_dictionary):
#     number_of_qubits = len(list(list(counts_dictionary.keys())[0]))
#     probability_distribution = np.zeros((2 ** number_of_qubits), dtype=float)
#     normalization_shots = sum(list(counts_dictionary.values()))

#     for key, shots in counts_dictionary.items():
#         probability_distribution[int(key, 2)] += shots / normalization_shots
#     return probability_distribution

# """

# """
# def get_qubits_to_clusters_map(clusters_division):
#     qubits_indices = []

#     for cluster in clusters_division:
#         qubits_indices += list(cluster)

#     qubits_to_clusters_map = {}
#     for qubit in qubits_indices:
#         for cluster in clusters_division:
#             if qubit in cluster:
#                 qubits_to_clusters_map[qubit] = cluster
#     return qubits_to_clusters_map
#     # #MOVE_TO >> core.quantum or to characterisation
#     # #DUPLICATE in dunctions data analysis.py
#     # def get_qubits_to_clusters_map(clusters_list,
#     #                             number_of_qubits):
#     #     qubits_to_clusters_map = {}
#     #     for qubit in range(number_of_qubits):
#     #         for cluster in clusters_list:
#     #             if qubit in cluster:
#     #                 qubits_to_clusters_map[qubit] = cluster

#     #     return qubits_to_clusters_map



# def get_pairs_complements_clusters(clusters_list):
#     pairs_complements_dictionary = {}
#     for cluster in clusters_list:
#         pairs_in_cluster = [(qi, qj) for qi in cluster for qj in cluster if qj > qi]

#         for pair in pairs_in_cluster:
#             pairs_complements_dictionary[pair] = utils.lists_difference(cluster, pair)
#     return pairs_complements_dictionary



# def divide_counts_dictionary_into_batches(counts_dictionary:Dict[str,int],
#                                           number_of_batches:int):
#     if number_of_batches ==1 or number_of_batches is None or number_of_batches==0:
#         return [counts_dictionary]

#     list_of_counts = []
#     for output_bitstring, number_of_ticks in counts_dictionary.items():
#         list_of_counts = list_of_counts + [output_bitstring] * number_of_ticks

#     rng = np.random.default_rng()
#     rng.shuffle(list_of_counts)
#     rng.shuffle(list_of_counts)
#     rng.shuffle(list_of_counts)
#     rng.shuffle(list_of_counts)
#     rng.shuffle(list_of_counts)

#     batch_size = len(list_of_counts) // number_of_batches
#     batches = [list_of_counts[index_batch * batch_size:(index_batch + 1) * batch_size]
#                for index_batch in range(number_of_batches)]

#     batched_dictionaries = [dict(Counter(batch_now)) for batch_now in batches]

#     return batched_dictionaries

# def divide_dot_results_dictionary_into_batches(dot_results:Dict[str,Dict[str,int]],
#                                                number_of_batches:int):
#     batched_results = [{} for _ in range(number_of_batches)]

#     for input_state, counts_dictionary in tqdm(dot_results.items()):
#         batched_counts = divide_counts_dictionary_into_batches(counts_dictionary=counts_dictionary,
#                                                                    number_of_batches=number_of_batches)

#         for batch_index in range(len(batched_counts)):
#             batched_results[batch_index][input_state] = batched_counts[batch_index]

#     return batched_results

# def version_control_results_dot(dictionary_data_counts:Dict[str,dict]):

#     try:
#         # Old convention, shouldn't be used anymore
#         dictionary_results = dictionary_data_counts['results_dictionary_preprocessed']

#     except(KeyError):

#         try:
#             dictionary_results = dictionary_data_counts['results_dictionary_preprocessed']
#         except(KeyError):
#             try:
#                 dictionary_results_DDOT = dictionary_data_counts['DDOT_results']

#             except(KeyError):
#                 dictionary_results_DDOT = {}



#             try:
#                 dictionary_results_QDOT = dictionary_data_counts['QDOT_results']
#             except(KeyError):
#                 dictionary_results_QDOT = {}

#             dictionary_results = {**dictionary_results_QDOT, **dictionary_results_DDOT}

#     return dictionary_results
# """
# class KeyDependentDictForMarginals(defaultdict):
#     # TODO FBM: This is too slow for big systems, remove usage of this from repository
#     """
#     This is class used to store marginal probability distributions in dictionary.
#     It is "key dependent" in the sense that if user tries to refer to non-existing value for some
#     KEY, then this value is created as a marginal distribution which size depends on the KEY
#     NOTE: We assume that provided KEY is a string denoting  qubits subset
#     (see self.value_creating_function)


#     COPYRIGHT NOTE
#     The main idea of this code was taken from Reddit thread:
#     https://www.reddit.com/r/Python/comments/27crqg/making_defaultdict_create_defaults_that_are_a/

#     """

#     def __init__(self):
#         super().__init__(None)  # initialize as standard defaultdict

#         # This is the function which takes the string "key" that is assumed to label qubits subset
#         # in the form 'q2q3q11...' etc. It takes this key, calculates number of qubits N, and creates
#         # empty vector of the size d=2^N.
#         self.value_creating_function = lambda key: np.zeros(
#             (int(2 ** len(key)), 1),
#             dtype=float)

#     # called when key is missing
#     def __missing__(self, key):
#         # calculate the key-dependent value
#         ret = self.value_creating_function(key)
#         # put the value inside the dictionary
#         self[key] = ret
#         return ret


# class key_dependent_dict_for_marginals(defaultdict):
#     """
#     same as KeyDependentDictForMarginals but different name
#     TODO FBM: refactor this
#     """

#     def __init__(self):
#         super().__init__(None)  # initialize as standard defaultdict

#         # This is the function which takes the string "key" that is assumed to label qubits subset
#         # in the form 'q2q3q11...' etc. It takes this key, calculates number of qubits N, and creates
#         # empty vector of the size d=2^N.
#         self.value_creating_function = lambda key: np.zeros(
#             (int(2 ** len(convert.get_qubit_indices_from_keystring(key))), 1),
#             dtype=float)

#     # called when key is missing
#     def __missing__(self, key):
#         # calculate the key-dependent value
#         ret = self.value_creating_function(key)
#         # put the value inside the dictionary
#         self[key] = ret
#         return ret


# """
# def get_state_from_circuit_name(circuit_name):
#     state_name = ''
#     for string in circuit_name:
#         if string in ['1', 'X', 'x']:
#             state_name += '1'
#         elif string in ['0', 'I', 'i_index', 'id', 'Id']:
#             state_name += '0'

#     return state_name


# def get_mini_dict(number_of_qubits):
#     register = povmtools.register_names_qubits(range(number_of_qubits), number_of_qubits)
#     return {key: np.zeros((int(2 ** number_of_qubits), 1)) for key in register}
# """



"""
Created on Wed Nov 22 23:50:46 2017

@author: Flavio Baccari
"""
from __future__ import print_function, division

import copy as cp
from operator import itemgetter
from time import time

import numpy as np
from qrem.qrem.ctmp.modeltools.ncpol2sdpa import flatten, SdpRelaxation, get_monomials
from qrem.qrem.ctmp.modeltools.ncpol2sdpa.nc_utils import get_support
from numpy import dot


def get_triang_ineqs(spins):
    """Generate the list of triangle inequalities involving the given spin
    variables

    :param spins: list of spin variables
    :type spins: list of sympy.core.symbol.Symbol
    :returns: list of sympy.core.add.Add
    """

    ineq = []

    # Run over all the triples
    for i in range(len(spins) - 2):
        for j in range(i + 1, len(spins) - 1):
            for k in range(j + 1, len(spins)):
                # Add all the different sign choices
                ineq.append(spins[i] * spins[j] + spins[i] * spins[k] +
                            spins[j] * spins[k] + 1)
                ineq.append(spins[i] * spins[j] - spins[i] * spins[k] -
                            spins[j] * spins[k] + 1)
                ineq.append(-spins[i] * spins[j] - spins[i] * spins[k] +
                            spins[j] * spins[k] + 1)
                ineq.append(-spins[i] * spins[j] + spins[i] * spins[k] -
                            spins[j] * spins[k] + 1)
    return ineq


def get_det_guess(s_variables, cliques, Mom, one, two):
    """Obtains the closest spin configuration from the moment matrix resulting from
    the lower bound SDP

    :param s_variables: list of spin variables
    :type s_variables: list of sympy.core.symbol.Symbol
    :param cliques: list of the cliques of the chordal extension
    :type cliques: list of lists of sympy.core.symbol.Symbol
    :param Mom: the block moment matrix at level 1
    :type Mom: list of numpy.ndarray
    :param one: coefficients of the one-body terms in the hamiltonian
    :type one: list of float
    :param two: coefficients of the two-body terms in the hamiltonian as a matrix
    :type one: list of list of float


    :returns: the deterministic configuration final_config as list of floats
    """

    # taking a Choleski decomposition of each block in the moment matrix
    Red = []
    configs = []

    for c, M in enumerate(Mom):

        B = np.linalg.cholesky(M + (10 ** (-6)) * np.identity(len(M)))
        BT = B.T.conj()
        # Get the smallest size of the vectors

        diff = np.linalg.norm(M - np.dot(B, BT))
        l = len(flatten(s_variables))

        while diff < 10 ** (-5):
            l -= 1
            tmp = B[::, 0:l]
            diff = np.linalg.norm(M - np.dot(tmp, tmp.T.conj()))

        d = l + 1
        Red.append(B[::, 0:d])

        configs.append([np.sign(dot(B[0, 0:d], el)) for el in B[1::, 0:d]])

    # Extract the overall configuration
    sub_conf = {}

    for b, block in enumerate(cliques):
        for e, el in enumerate(block):
            sub_conf[el] = configs[b][e]

    config = []

    for el in flatten(s_variables):
        config.append(sub_conf[el])

    bound = get_bound_from_det(flatten(config), one, two)

    final_config = cp.copy(config)
    # Check if the bound can be improved by flipping a single spin

    for i in range(len(config)):

        flip_config = cp.deepcopy(config)
        flip_config[i] = cp.deepcopy(-flip_config[i])

        bound_tmp = get_bound_from_det(flatten(flip_config), one, two)

        if bound_tmp < bound:
            bound = cp.deepcopy(bound_tmp)
            final_config = cp.deepcopy(flip_config)

    return final_config


def get_bound_from_det(config, one, two):
    """Gets the energy of the corresponding configuration

    :param config: the spin configuration
    :type config: list of float
    :param one: coefficients of the one-body terms in the hamiltonian
    :type one: list of float
    :param two: coefficients of the two-body terms in the hamiltonian as a matrix
    :type one: list of list of float


    :returns: the upper bound to the ground state energy as float
    """
    # Config has to be flattened

    bound = 0

    # One-body
    for i, el in enumerate(config):
        bound += one[i] * el

    # Two-body

    for i, el1 in enumerate(config):
        for j, el2 in enumerate(config[i + 1::], i + 1):
            bound += two[i][j] * el1 * el2

    return bound


def get_strengthen_low(s_variables, substitutions, hamiltonian, cliques,
                       threshold, solverparameters=None, verbose=0):
    """Runs the relaxation for the first time and determines how many triangle inequalities
    to add to get a reasonably good lower bound. Outputs the corresponding lower bound
    as well

    :param s_variables: list of spin variables
    :type s_variables: list of sympy.core.symbol.Symbol
    :param substitutions: substitution to be applied in the generation of the moment matrix at the level of operators
    :type substitutions: dict of items of polynomials of sympy.core.symbol.Symbol
    :param hamiltonian: hamiltonian of the problem as a symbolic polynomial of the spins
    :type hamiltonian: sympy.core.add.Add
    :param cliques: list of the cliques of the chordal extension
    :type cliques: list of lists of sympy.core.symbol.Symbol
    :param threshold: threshold clique size n_t under which generats blocks at level 2
    :type threshold: int
    :param solverparameters: parameters for the SDP solver
    :type solverparameters: dict
    :param verbose: verbosity level
    :type verbose: int

    :returns: sdp as ncpol2sdpa sdp class, list of triangle inequalities to add as
    list of sympy.core.add.Add
    """
    N = len(flatten(s_variables))
    # n_ineqs = N
    new_ineqs = []
    ineqs = []
    # Generate the mixed level with blocks al level 2 only if the are below
    # the threshold size
    monomial_blocks = []

    for el in cliques:

        if len(el) <= threshold:

            monomial_blocks.append(get_monomials(el, 2))
        else:
            monomial_blocks.append(get_monomials(el, 1))

    # Solve the first SDP

    sdp = SdpRelaxation(flatten(s_variables), verbose=verbose)
    sdp.get_relaxation(
        -1,
        substitutions=substitutions,
        objective=hamiltonian,
        momentinequalities=flatten(ineqs),
        extramonomials=monomial_blocks)

    sdp.solve(solver="mosek", solverparameters=solverparameters)

    bound_init = cp.deepcopy(sdp.primal)
    bound_new = cp.deepcopy(sdp.primal)

    # Get the moment matrix without the identity and computing the triangular
    # inequalities

    lengths = [len(el) for el in cliques]
    improv = 1

    while (improv > 0.005):

        bound_old = cp.deepcopy(sdp.primal)
        ineqs_values = []

        for c in range(len(cliques)):

            M = cp.deepcopy(sdp.x_mat[c][1:lengths[c] + 1, 1:lengths[c] + 1])

            for i in range(lengths[c] - 2):
                for j in range(i + 1, lengths[c] - 1):
                    for k in range(j + 1, lengths[c]):
                        tmp = []

                        tmp.append(M[i, j] + M[j, k] + M[i, k] + 1)
                        tmp.append(M[i, j] - M[j, k] - M[i, k] + 1)
                        tmp.append(-M[i, j] + M[j, k] - M[i, k] + 1)
                        tmp.append(-M[i, j] - M[j, k] + M[i, k] + 1)

                        ineqs_values.append([[c, i, j, k], min(tmp)])

        # Select the first N most violated ineqs and add the corresponding 4
        # constraints to the problem
        # print(ineqs_values)
        # print([el[1] for el in ineqs_values])
        # print()

        n_ineqs = min([len(ineqs_values), N])

        for _ in range(n_ineqs):
            b = [el[1] for el in ineqs_values]
            # print(ineqs_values)
            index = min(enumerate(b), key=itemgetter(1))[0]
            [c, i, j, k] = ineqs_values[index][0]
            ineqs_values.remove(ineqs_values[index])

            variables = [
                flatten(cliques[c])[i],
                flatten(cliques[c])[j],
                flatten(cliques[c])[k]
            ]
            new_ineqs.append(get_triang_ineqs(variables))

        # Solve the new sdp

        time0 = time()

        sdp = SdpRelaxation(flatten(s_variables), verbose=verbose)
        sdp.get_relaxation(
            -1,
            substitutions=substitutions,
            objective=hamiltonian,
            momentinequalities=flatten(new_ineqs),
            extramonomials=monomial_blocks)

        sdp.solve(solver="mosek", solverparameters=solverparameters)

        bound_new = cp.deepcopy(sdp.primal)
        improv = abs(bound_new - bound_old) / abs(bound_init)

        # if the relative improvement is good enough add the new inequalities

        if (improv > 0.005):
            ineqs = cp.deepcopy(new_ineqs)

    return sdp, ineqs


def get_Mom(sdp, cliques):
    """Extract the level 1 moment matrix from the solved sdp problem

    :param spd: SDP problem generated by ncpol2sdpa
    :type sdp: ncpol2sdpa.sdp_relaxation.SdpRelaxation
    :param cliques: list of the cliques of the chordal extension
    :type cliques: list of lists of sympy.core.symbol.Symbol


    :returns: the level 1 the block moment matrix Mom as list of numpy.ndarray,
    the average spin values as list of numpy.ndarray
    """

    # Extract the moment matrix for the cliques

    lengths = [len(el) for el in cliques]
    Mom = []
    spins = []

    for c in range(len(cliques)):
        M = cp.copy(sdp.x_mat[c][0:lengths[c] + 1, 0:lengths[c] + 1])
        Mom.append(sdp.x_mat[c][0:lengths[c] + 1, 0:lengths[c] + 1])

        spins.append(M[1:, 0])

    return Mom, spins


def get_up_and_low(s_variables, substitutions, ineqs, hamiltonian, one, two,
                   cliques, threshold, solverparameters=None, verbose=0):
    """Gets upper and lower bound for the given instance of branching

    :param s_variables: list of spin variables
    :type s_variables: list of sympy.core.symbol.Symbol
    :param substitutions: substitution to be applied in the generation of the moment matrix at the level of operators
    :type substitutions: dict of items of polynomials of sympy.core.symbol.Symbol
    :param ineqs: triangle inequalities to be added to improve the lower bound
    :type ineqs: list of sympy.core.add.Add
    :param hamiltonian: hamiltonian of the problem as a symbolic polynomial of the spins
    :type hamiltonian: sympy.core.add.Add
    :param one: coefficients for the one-body terms in the hamiltonian
    :type one: list of floats
    :param two: coefficients for the two-body terms in the hamiltonian
    :type two: list of floats
    :param cliques: list of the cliques of the chordal extension
    :type cliques: list of lists of sympy.core.symbol.Symbol
    :param threshold: threshold clique size n_t under which generats blocks at level 2
    :type threshold: int
    :param solverparameters: parameters for the SDP solver
    :type solverparameters: dict
    :param verbose: verbosity level
    :type verbose: int

    :returns: lower bound as float, upper bound as float, spin expectation values
    extracted from the moment matrix for each clique as list of array of floats,
    the deterministic configuration for each spin as list of floats
    """

    time0 = time()

    monomial_blocks = []

    for el in cliques:

        if len(el) <= threshold:

            monomial_blocks.append(get_monomials(el, 2))
        else:
            monomial_blocks.append(get_monomials(el, 1))

    # solving the first sdp

    sdp = SdpRelaxation(flatten(s_variables), verbose=verbose)
    sdp.get_relaxation(
        -1,
        substitutions=substitutions,
        objective=hamiltonian,
        momentinequalities=flatten(ineqs),
        extramonomials=monomial_blocks)

    sdp.solve(solver="mosek", solverparameters=solverparameters)

    print("The lower bound is " + str(sdp.primal))
    print("Time taken for lower is " + str(time() - time0))

    time0 = time()
    [Mom, spins] = get_Mom(sdp, cliques)

    config = get_det_guess(s_variables, cliques, Mom, one, two)
    bound = get_bound_from_det(flatten(config), one, two)
    print("Time taken for upper is " + str(time() - time0))

    return sdp.primal, bound, spins, config


def get_branching_node(spins, s_variables, cliques, eqs):
    """Gets the spin on which to perform the next branching. Works with an "easy-first"
    choice

    :param spins: the spin average values extracted from the moment matrix
    :type spins: list of floats
    :param s_variables: list of spin variables
    :type s_variables: list of sympy.core.symbol.Symbol
    :param cliques: list of the cliques of the chordal extension
    :type cliques: list of lists of sympy.core.symbol.Symbol
    :param eqs: list of already implemented branching choices
    :type eqs: list of sympy.core.add.Add

    :returns: the spin variable of the branching node as sympy.core.symbol.Symbol
    """

    det = 1
    sup = get_sup(s_variables, eqs)

    for c, block in enumerate(spins):
        for i, el in enumerate(block):

            if el not in sup:

                if abs(el) < det:
                    det = cp.deepcopy(abs(el))
                    branch_node = cp.deepcopy(cliques[c][i])

    return branch_node


def get_as_coeffs(hamiltonian, s_variables):
    """Extracts the one and two-body coefficients from the symbolic hamiltonian

    :param hamiltonian: hamiltonian of the problem as a symbolic polynomial of the spins
    :type hamiltonian: sympy.core.add.Add
    :param s_variables: list of spin variables
    :type s_variables: list of sympy.core.symbol.Symbol

    :returns: the one-body coeffs as list of float, the two-body coeffs as list
    of list of float
    """

    one = []
    two = []
    pol = hamiltonian.as_poly()
    mon = flatten(s_variables)

    # print(pol)

    for i, el in enumerate(mon):

        one.append(pol.coeff_monomial(el))
        tmp = []
        for j, el2 in enumerate(mon):
            if j < i:
                tmp.append(0)
            else:

                # print(el, el2)
                tmp.append(pol.coeff_monomial(el * el2))
        two.append(tmp)

    return one, two


def get_sup(s_variables, eqs):
    """Gets the spin variables on which the branching has alreadu been performed

    :param s_variables: list of spin variables
    :type s_variables: list of sympy.core.symbol.Symbol
    :param eqs: list of already implemented branching choices
    :type eqs: list of sympy.core.add.Add

    :returns: support of the branching equations as list of sympy.core.symbol.Symbol
    """

    sup = []

    for el in eqs:
        s = get_support(flatten(s_variables), el)
        sup.append(dot(s[1], flatten(s_variables)))

    return sup



def get_groundBandB(s_variables,
                    substitutions,
                    hamiltonian,
                    cliques,
                    threshold=3,
                    verbose=0,
                    # solverparameters=None
                    ):
    """Runs the whole branch&bound algorithm for the given hamiltonian

    :param s_variables: list of spin variables
    :type s_variables: list of sympy.core.symbol.Symbol
    :param substitutions: substitution to be applied in the generation of the moment matrix at the level of operators
    :type substitutions: dict of items of polynomials of sympy.core.symbol.Symbol
    :param hamiltonian: hamiltonian of the problem as a symbolic polynomial of the spins
    :type hamiltonian: sympy.core.add.Add
    :param cliques: list of the cliques of the chordal extension
    :type cliques: list of lists of sympy.core.symbol.Symbol
    :param threshold: threshold clique size n_t under which generats blocks at level 2
    :type threshold: int
    :param solverparameters: parameters for the SDP solver
    :type solverparameters: dict
    :param verbose: verbosity level
    :type verbose: int

    :returns z_low: final lower bound as float
    :returns z_up: final upper bound as float
    :returns final_config: final ground state configuration as list of floats
    """

    [one, two] = get_as_coeffs(hamiltonian, s_variables)
    # First root

    eqs = []
    tree = []
    print("Starting initial optimisation")

    # Get the first lower bound and choosing the list of triangle ineqs to add

    [sdp, ineqs] = get_strengthen_low(
        s_variables,
        substitutions,
        hamiltonian,
        cliques,
        threshold,
        # solverparameters,
        verbose=0)

    # Extract the corresponding upper bound and spin configuration

    [Mom, spins] = get_Mom(sdp, cliques)
    config = get_det_guess(s_variables, cliques, Mom, one, two)
    final_config = cp.deepcopy(config)

    z_up = get_bound_from_det(flatten(config), one, two)
    z_low = sdp.primal

    print("The lower bound is " + str(z_low))
    print("The upper bound is " + str(z_up))

    # Go for the branching with easy first

    if ((z_low - z_up) / abs(z_up) < -0.00001):
        node = get_branching_node(spins, s_variables, cliques, eqs)
        tree = [[z_low, eqs, substitutions, node]]

    while (len(tree) > 0):

        if abs(z_up-z_low)<10**(-6):
            #TODO FBM: added this
            print('bounds converged, returning solution')
            break

        lis = [el[0] for el in tree]
        index = min(enumerate(lis), key=itemgetter(1))[0]
        [z_low, eqs, subs, node] = tree[index]
        tree.remove(tree[index])

        # Branch +1

        eqs_new = cp.deepcopy(eqs)
        eqs_new.append(node - 1)
        subs_new = cp.deepcopy(subs)
        subs_new[node] = 1

        [z_low, tmp, spins, config] = get_up_and_low(
            s_variables,
            subs_new,
            ineqs,
            hamiltonian,
            one,
            two,
            cliques,
            threshold,
            # solverparameters,
            verbose=0)

        # Substitute the upper bound if better
        if tmp < z_up:
            z_up = cp.deepcopy(tmp)
            final_config = cp.deepcopy(config)

        # Include this node in the branching if the lower bound is meaningful

        if ((z_low - z_up) / abs(z_up) < -0.00001):
            node_new = get_branching_node(spins, s_variables, cliques, eqs_new)
            tree.append([z_low, eqs_new, subs_new, node_new])

        # branch -1

        eqs_new = cp.deepcopy(eqs)
        eqs_new.append(node + 1)
        subs_new = cp.deepcopy(subs)
        subs_new[node] = -1

        [z_low, tmp, spins, config] = get_up_and_low(
            s_variables,
            subs_new,
            ineqs,
            hamiltonian,
            one,
            two,
            cliques,
            threshold,
            # solverparameters,
            verbose=verbose)

        # Substitute the upper bound if better
        if tmp < z_up:
            z_up = cp.deepcopy(tmp)
            final_config = cp.deepcopy(config)

        # Include this node in the branching if the lower bound is meaningful
        print("The upper bound is " + str(z_up))
        if ((z_low - z_up) / abs(z_up) < -0.00001):
            node_new = get_branching_node(spins, s_variables, cliques, eqs_new)
            tree.append([z_low, eqs_new, subs_new, node_new])

        # Check all the previous nodes to see if we can eliminate some with
        # the new z_up

        tree_old = cp.deepcopy(tree)
        tree = []

        for el in tree_old:

            if ((el[0] - z_up) / abs(z_up) < -0.00001):
                tree.append(el)

    return z_low, z_up, final_config



import copy
from typing import Optional, Dict, Union, Tuple

import numpy as np
import scipy as sc
from tqdm import tqdm

from qrem.functions_qrem import ancillary_functions as anf
from qrem.noise_characterization.tomography_design.overlapping.DOTMarginalsAnalyzer import DOTMarginalsAnalyzer
from qrem.noise_model_generation.CTMP import functions_CTMP as fun_CTMP

import qrem.common.utils as qrem_utils
# from qrem.common.printer import qprint_array, qprint

class ModelGeneratorCTMP(DOTMarginalsAnalyzer):

    # TODO Significant refactoring needed
    #

    def __init__(self,
                 results_dictionary_ddot: Dict[str, Dict[str, int]],

                 number_of_qubits: int,
                 bitstrings_right_to_left: Optional[bool] = False,
                 marginals_dictionary: Optional[Dict[str, Dict[str, np.ndarray]]] = None,
                 noise_matrices_dictionary: Optional[
                     Dict[str, Union[np.ndarray, Dict[str, Dict[str, np.ndarray]]]]] = None,
                 pairs_complements_dictionary=None
                 ) -> None:
        """

        :param results_dictionary_ddot:
        :type results_dictionary_ddot:
        :param bitstrings_right_to_left:
        :type bitstrings_right_to_left:
        :param number_of_qubits:
        :type number_of_qubits:
        :param marginals_dictionary:
        :type marginals_dictionary:
        :param noise_matrices_dictionary:
        :type noise_matrices_dictionary:

        :param pairs_complements_dictionary:
        Dictionary with structure:

        key: pair tuple
        value: list/tuple of qubits which are correlated to it

        For given clusters' structure, this can be generated via function:
        noise_model_generation.CTMP.fucntions_CTMP.get_pairs_complement_dictionary_from_clusters_list


        :type pairs_complements_dictionary:
        """

        super().__init__(results_dictionary_ddot,
                         bitstrings_right_to_left,
                         marginals_dictionary,
                         noise_matrices_dictionary
                         )
        self._number_of_qubits = number_of_qubits
        self._qubit_indices = list(range(number_of_qubits))

        self._correlations_table_pairs = None
        self._rates_dictionary = {}

        if pairs_complements_dictionary is None:
            pairs_complements_dictionary = {}
            single_qubits = list(range(self._number_of_qubits))
            for qi in range(self._number_of_qubits):
                for qj in range(qi + 1, self._number_of_qubits):
                    pair = (qi, qj)
                    pair_complement = sorted(list(set(single_qubits).difference(set(pair))))
                    pairs_complements_dictionary[pair] = pair_complement

        self._pairs_complements_dictionary = pairs_complements_dictionary
        self._noise_strength_CTMP = None

        if noise_matrices_dictionary is None:
            # single_qubits = list(range(number_of_qubits))
            # pairs = [(i, j) for i in range(number_of_qubits) for j in range(i + 1, number_of_qubits)]

            self._noise_matrices_dictionary = {}

        else:
            self._noise_matrices_dictionary = noise_matrices_dictionary

    @property
    def rates_dictionary(self) -> Dict[Tuple[int], Dict[str, float]]:
        return self._rates_dictionary

    @rates_dictionary.setter
    def rates_dictionary(self, rates_dictionary: Dict[Tuple[int], Dict[str, float]]) -> None:
        self._rates_dictionary = rates_dictionary


    def find_optimal_noise_strength_CTMP(self):
        method_searching = "SDP_relaxation"
        if self._number_of_qubits<=22:
            method_searching='bruteforce'
        self._noise_strength_CTMP = fun_CTMP.find_optimal_noise_strength_CTMP(ctmp_rates_dictionary=self._rates_dictionary,
                                                                              number_of_qubits=self._number_of_qubits,
                                                                              method_name=method_searching)
        return self._noise_strength_CTMP


    def _get_local_noise_matrix_postselection(self,
                                              pair):

        # single_qubits = list(range(self._number_of_qubits))
        # print(pair)
        pair_complement = self._pairs_complements_dictionary[pair]
        # print(pair_complement)

        local_noise_matrix = np.zeros((4,4), dtype=float)
        for global_input_state, counts_dict in self._results_dictionary.items():
            marginal_input_state = ''.join([global_input_state[x] for x in pair])
            input_state_complement = [global_input_state[x] for x in pair_complement]

            # print(global_input_state)
            for global_output_state, ticks in counts_dict.items():
                marginal_output_state = ''.join([global_output_state[x] for x in pair])
                output_state_complement = [global_output_state[x] for x in pair_complement]

                # print(global_output_state)
                if output_state_complement == input_state_complement:
                    # print(output_state_complement,input_state_complement,ticks)
                    # check if this is their convention!

                    local_noise_matrix[int(marginal_output_state, 2),
                                       int(marginal_input_state, 2)] += ticks
        # raise KeyboardInterrupt
        # qprint_array(local_noise_matrix)
        # normalize to stochastic matrix
        for k in range(4):
            sum_column = sum(local_noise_matrix[:, k])
            if sum_column==0:
                local_noise_matrix[k,k] = 1
            else:
                local_noise_matrix[:, k] /= sum(local_noise_matrix[:, k])

        # qprint_array(local_noise_matrix)
        # raise KeyboardInterrupt
        return local_noise_matrix

    def _get_local_noise_matrix_postselection_1q(self,
                                                 qubit):

        local_noise_matrix = np.zeros((2, 2), dtype=float)
        for global_input_state, results_dictionary in self._results_dictionary.items():
            marginal_input_state = global_input_state[qubit]

            for global_output_state, ticks in results_dictionary.items():
                marginal_output_state = global_output_state[qubit]

                local_noise_matrix[int(marginal_output_state, 2),
                                   int(marginal_input_state, 2)] += ticks
        # normalize to stochastic matrix
        for k in range(2):
            local_noise_matrix[:, k] /= sum(local_noise_matrix[:, k])

        return local_noise_matrix

    def _get_local_noise_matrix(self,
                                pair,
                                method='postselection'):

        if method.upper() == 'POSTSELECTION':
            return self._get_local_noise_matrix_postselection(pair=pair)
        else:
            raise ValueError(f"Method '{method}' not yet supported")

    def _calculate_multiple_noise_matrices(self,
                                           pairs_list,
                                           method='postselection'):
        pairs_range = copy.deepcopy(pairs_list)
        if len(pairs_list)>50:
            pairs_range = tqdm(pairs_range)
            # print(pairs_list)

        for pair in pairs_range:
            local_noise_matrix = self._get_local_noise_matrix(pair=pair,
                                                              method=method)
            # print(local_noise_matrix,'hej')

            self._noise_matrices_dictionary[pair] = local_noise_matrix


    # @staticmethod
    # def _get_1q_G_matrix(stochastic_map):
    #
    #     epsilon = stochastic_map[1,0]
    #     eta = stochastic_map[0,1]
    #
    #     G_matrix = -np.log(1-epsilon-eta)/(epsilon)
    #

    def _get_CTMP_rates_from_results_postselection(self,
                                                   pairs_list=None
                                                   ):

        number_of_qubits = self._number_of_qubits
        single_qubits = self._qubit_indices

        if pairs_list is None:
            pairs_list = [(qi, qj) for qi in single_qubits for qj in single_qubits if qj > qi]

        # print(pairs_list)


        qubits_in_pairs = []
        for pair in pairs_list:
            for qubit in pair:
                qubits_in_pairs.append(qubit)
        unique_qubits_in_pairs = sorted(list(np.unique(qubits_in_pairs)))
        unique_qubits_alone = qrem_utils.lists_difference(list(range(number_of_qubits)),
                                                   unique_qubits_in_pairs)




        self._calculate_multiple_noise_matrices(pairs_list=pairs_list,
                                                method='postselection')

        # Get G matrices
        G_matrices = {pair: sc.linalg.logm(self._noise_matrices_dictionary[pair]) for pair in
                      pairs_list}

        # ancillary function
        def __chop_negatives(M):
            (m, n) = M.shape

            chopped_M = copy.deepcopy(M)
            for i in range(m):
                for j in range(n):
                    if i != j and M[i, j] < 0:
                        chopped_M[i, j] = 0

            return chopped_M

        # Get G' matrices
        G_prime_matrices = {pair: __chop_negatives(G_matrices[pair]) for pair in pairs_list}


        rates_dictionary_1q_new_format = {(q,): {'0': 0, '1': 0} for q in single_qubits}


        for qj in unique_qubits_in_pairs:
            # qprint("hey",qj)
            r0, r1 = 0, 0

            counter_normalization = 0
            for q_other in unique_qubits_in_pairs:
                if q_other != qj:
                    # qprint("hey", q_other,'red')

                    if tuple(sorted([qj, q_other])) in G_prime_matrices.keys():
                        # try:
                        G_prime_matrix_now = G_prime_matrices[tuple(sorted([qj, q_other]))]

                        if qj<q_other:

                            r0 += G_prime_matrix_now[2, 0] + G_prime_matrix_now[3, 1]
                            r1 += G_prime_matrix_now[0, 2] + G_prime_matrix_now[1, 3]
                        else:
                            r0 += +G_prime_matrix_now[1,0]+G_prime_matrix_now[3,2]
                            r1 += G_prime_matrix_now[0, 1] + G_prime_matrix_now[2, 3]

                        counter_normalization+=2
                        # except(KeyError):
                            #     pass


            r0 /= counter_normalization
            r1 /= counter_normalization

            rates_dictionary_1q_new_format[(qj,)]['0'] = r0
            rates_dictionary_1q_new_format[(qj,)]['1'] = r1


        # for averaged_1q matrix
        G_matrices_1q = {
            (qi,): sc.linalg.logm(self._get_local_noise_matrix_postselection_1q(qubit=qi))
            for qi in unique_qubits_alone}
        G_prime_matrices_1q = {qi: __chop_negatives(G_matrix) for qi, G_matrix in
                               G_matrices_1q.items()}



        for qi, G_prime_1q in G_prime_matrices_1q.items():
            r0 = G_prime_1q[1, 0]
            r1 = G_prime_1q[0, 1]

            rates_dictionary_1q_new_format[qi]['0'] = r0
            rates_dictionary_1q_new_format[qi]['1'] = r1



        rates_dictionary_2q_new_format = {pair: {'00': 0,
                                                 '01': 0,
                                                 '10': 0,
                                                 '11': 0} for pair in pairs_list}

        for pair in pairs_list:
            G_prime_matrix_now = G_prime_matrices[pair]

            rates_dictionary_2q_new_format[pair]['00'] = G_prime_matrix_now[3, 0]
            rates_dictionary_2q_new_format[pair]['01'] = G_prime_matrix_now[2, 1]
            rates_dictionary_2q_new_format[pair]['10'] = G_prime_matrix_now[1, 2]
            rates_dictionary_2q_new_format[pair]['11'] = G_prime_matrix_now[0, 3]

        rates_dictionary_new_format = {**rates_dictionary_1q_new_format,
                                       **rates_dictionary_2q_new_format}

        return rates_dictionary_new_format

    def get_CTMP_rates_from_results(self,
                                    pairs_list=None,
                                    method='postselection'):

        if method.upper() == 'POSTSELECTION':
            rates_dictionary = self._get_CTMP_rates_from_results_postselection(pairs_list=pairs_list)
        else:
            raise ValueError(f"Method '{method}' is not yet supported.")

        self._rates_dictionary = rates_dictionary

        return self._rates_dictionary







"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com
"""
import qutip
import numpy as np
from typing import Optional, Union, List, Tuple,Dict
from qrem.functions_qrem import ancillary_functions as anf

def _embed_1q_operator(number_of_qubits:int,
                       local_operator:np.ndarray,
                       global_qubit_index:int):
    if global_qubit_index == 0:
        embed_operator = np.kron(local_operator, np.eye(int(2 ** (number_of_qubits - 1))))
        return embed_operator
    else:
        first_eye = np.eye(2 ** (global_qubit_index))
        second_eye = np.eye(2 ** (number_of_qubits - global_qubit_index - 1))

        embeded_operator = np.kron(np.kron(first_eye, local_operator), second_eye)

        return embeded_operator


def embed_operator_in_bigger_hilbert_space(number_of_qubits:int,
                                           local_operator:np.ndarray,
                                           global_indices:Optional[Union[List[int],Tuple[int]]]=[0, 1]):

    N_small = int(np.log2(local_operator.shape[0]))

    # print
    if N_small == 1:
        return _embed_1q_operator(number_of_qubits, local_operator, global_indices[0])


    qutip_object = qutip.Qobj(local_operator,
                              dims=[[2 for _ in range(N_small)],
                                    [2 for _ in range(N_small)]])


    return qutip.qip.operations.gates.expand_operator(qutip_object,
                                     number_of_qubits,
                                     global_indices).full()


#(PP) TODO_MO This function is defined around 4-5 times, with differing implementations. We need to unpack it together and simplify
def get_energy_from_bitstring_diagonal(bitstring: List[str],
                                       weights_dict: Dict[Tuple[int], float]
                                       ):
    energy = 0
    for qubits_tuple, hamiltonian_coefficient in weights_dict.items():
        marginal_bitstring = [int(bitstring[q]) for q in qubits_tuple]
        parity = (-1)**(np.count_nonzero(marginal_bitstring))
        energy += parity * hamiltonian_coefficient

    return energy



#(PP) not used - commented out
# def get_plus_state(N):
#     return np.full((2**N,1),(1/np.sqrt(2)**N))


#(PP) not used - commented out
# def get_bistring_parity(bitstring:List[Union[int,str]]):
#     if isinstance(bitstring[0],str):
#         bitstring = [int(x) for x in bitstring]

#     return (-1) ** (np.count_nonzero(bitstring))


#(PP) not used - commented out
# def get_energy_from_spectrum_and_probability(spectrum:Union[np.ndarray,List[float]],
#                                              probability:Union[np.ndarray,List[float]]):
#     return sum([spectrum[i] * probability[i] for i in range(len(probability))])


"""
@authors: Filip Maciejewski, Oskar Słowik, Tomek Rybotycki
@contact: filip.b.maciejewski@gmail.com
"""
# ORGANIZE from this file only function print_partition is used in non-deprecated code (we can discuss deleting that use);
# print_partition also uses create_undirected_nxgraph_from_partition; other functions can be deleted
from itertools import chain

import matplotlib.pyplot as plt
import networkx as nx
from qrem.functions_qrem import ancillary_functions as anf

# ORGANIZE - out of the non deprecated/ for deletion functions this is only used below in print_partition
def create_undirected_nxgraph_from_partition(partition,
                                             vertices_weights=None,
                                             edges_weights=None):
    # if vertices_labels is None:
    # vertices_labels
    q_list = [q for C in partition for q in C]

    G = nx.Graph()
    for qi in q_list:
        if vertices_weights is None:
            G.add_node(qi)
        else:
            G.add_node(qi, weight=vertices_weights[qi])

    # G.add_nodes_from(q_list)
    for C in partition:
        for i in range(len(C) - 1):
            for j in chain(range(i), range(i + 1, len(C))):
                if edges_weights is None:
                    G.add_edge(C[i], C[j])
                else:
                    G.add_edge(C[i], C[j],
                               weight=edges_weights[(C[i], C[j])])
    return G

# ORGANIZE - delete, this function is not used anywhere
def print_graph(graph):
    nx.draw_circular(graph, with_labels=True)
    plt.show()

# ORGANIZE - this is only used in functions_noise_model_heuristic(_OS), when an the parameter 'drawing' is set to True
# prints clusters with edges only within a cluster
def print_partition(partition):
    G = create_undirected_nxgraph_from_partition(partition)
    nx.draw_circular(G, with_labels=True)
    plt.show()
    # plt.savefig("path.png")


def __combine_colors(bad_color,
                     good_color,
                     bad_weight):
    bad_weight = bad_weight ** (1 / 2.5)

    new_color = [bad_weight * bad_color[i] + (1 - bad_weight) * good_color[i] for i in
                 range(3)] + [1]

    return new_color

# ORGANIZE - to delete; this function is used nowhere; similar functionality implemented in method draw_clusters in
# visualisation.ResultsForPlotting
def create_clusters_graph(clusters_list,
                          correlations_table,
                          true_qubits,
                          Title='',
                          errors_1q=None,

                          ):
    number_of_qubits = len(true_qubits)
    import numpy as np
    C_max = int(np.max([len(x) for x in clusters_list]))

    qubits_indices_map = dict(enumerate(true_qubits))
    # qubits_indices_rev_map = utils.map_index_to_order(true_qubits)
    # colors_list = [[256 / 256 * (i / number_of_qubits),
    #                 256 / 256 * (1 - i / number_of_qubits),
    #                 0,
    #                 1] for i in range(number_of_qubits)]
    #
    # original_color = (220 / 256, 1, 0)

    from matplotlib import pyplot as plt
    from matplotlib import rc
    import networkx as nx

    # fontname= 'Comic Sans MS'
    fontname = 'fantasy'
    plt.rcParams.update({'font.family': fontname})
    rc('font', **{'family': fontname})
    rc('text', usetex=True)
    #
    import numpy as np

    bad_color = (220 / 255, 20 / 255, 60 / 255)

    bad_color = (1, 0, 0)
    good_color = (0, 1, 0)


    node_size = 4000

    if number_of_qubits>30:
        node_size*=1/4

    # print(colors_list)
    # raise KeyError

    vertices_weights_list = []
    vertices_weights_dict = {}

    edges_weights = {}

    for i in range(number_of_qubits):
        if errors_1q is None:
            err_1q_now = 0.1
        else:
            err_1q_now = (errors_1q['q%s' % i][0, 1] +
                               errors_1q['q%s' % i][1, 0]) / 2

        vertices_weights_dict[qubits_indices_map[i]] = err_1q_now
        vertices_weights_list.append((qubits_indices_map[i], err_1q_now))

        for j in range(number_of_qubits):
            edges_weights[(qubits_indices_map[i], qubits_indices_map[j])] \
                = 1 / 2 * (correlations_table[i, j] + correlations_table[j, i])

    # print(vertices_weights_dict)
    # raise KeyError
    # vertices_weights_list = sorted(vertices_weights_list, key=lambda x: x[1])
    # # print(vertices_weights_list)
    if errors_1q is None:
        weights_colors = {u:'cornflowerblue' for u, v in vertices_weights_list}

    else:
        weights_colors = {u: __combine_colors(bad_color=bad_color,
                                              good_color=good_color,
                                              bad_weight=v / np.max(
                                                 list(vertices_weights_dict.values())))
                          for u, v in vertices_weights_list}

    # print(weights_colors)
    # raise KeyError

    clusters_mapped = [[qubits_indices_map[qi] for qi in cluster] for cluster in clusters_list]
    graph = create_undirected_nxgraph_from_partition(clusters_mapped,
                                                     vertices_weights=vertices_weights_dict,
                                                     edges_weights=edges_weights)

    positions = {}

    if C_max == 5:
        x_pos, y_pos = 0, 1.5
    else:
        x_pos, y_pos = 0, 1

    clusters_mapped = sorted(clusters_mapped, key=lambda x: len(x))

    for cluster in clusters_mapped:
        graph_cluster = create_undirected_nxgraph_from_partition([cluster],
                                                                 edges_weights=edges_weights)

        graph_design = nx.drawing.layout.circular_layout(G=graph_cluster,
                                                         # k=0.3,

                                                         # pos=positions
                                                         )

        for qi, position_now in graph_design.items():
            positions[qi] = (
                position_now[0] / len(clusters_list) + x_pos, position_now[1] / 10 + y_pos)

        if C_max == '5':
            x_pos += 0.75
            if x_pos >= 1.5:
                x_pos -= 1.5

                y_pos -= 0.3

        else:
            x_pos += 0.3

            if x_pos >= 1:
                x_pos -= 1
                y_pos -= 0.3

    if C_max == '5':
        fig = plt.figure(figsize=(25, 25))
    else:
        fig = plt.figure(figsize=(15, 15))

    # Title = f"Cmax = {C_max}, version = {version}, alpha = {alpha}"

    plt.title(Title, fontdict={'fontsize': 60})
    nx.draw(graph,
            with_labels=True,
            pos=positions,
            node_size=node_size,
            font_size=40,
            width=[graph[u][v]['weight'] / np.max(list(edges_weights.values())) * 10 + 0.5 for
                   u, v in
                   graph.edges],
            edge_color=[0, 0, 0, 1],
            node_color=[weights_colors[qi] for qi in graph.nodes]
            # width
            )

    # figname = f"Cmax = {C_max}, version = {version}, {counter}_alpha = {alpha}"
    # directory_with_figures = directory_to_open + f'figures/clusters_visualization/C_max{C_max}/'
    plt.tight_layout()
    plt.show()



"""
@authors: Filip Maciejewski, Oskar Słowik, Tomek Rybotycki
@contact: filip.b.maciejewski@gmail.com

REFERENCES:
[0] Filip B. Maciejewski, Zoltán Zimborás, Michał Oszmaniec,
"Mitigation of readout noise in near-term quantum devices
by classical post-processing based on detector tomography",
Quantum 4, 257 (2020)

[1] Z. Hradil, J. Řeháček, J. Fiurášek, and M. Ježek, “3 maximum-likelihood methods in quantum mechanics,” in Quantum
State Estimation, edited by M. Paris and J. Řeháček (Springer Berlin Heidelberg, Berlin, Heidelberg, 2004) pp. 59–112.
[2] J. Fiurášek, Physical Review arrray_to_print 64, 024102 (2001), arXiv:quant-ph/0101027 [quant-ph].
"""
import copy
import time
from typing import List, Dict,Union,Optional

from math import log
import numpy as np
import scipy as sc
import qutip
from qiskit.result import Result

from qrem.common import utils
from qrem.functions_qrem.povmtools import get_density_matrix, permute_matrix, reorder_classical_register
from qrem.common.providers.ibmutils.data_converters import get_frequencies_array_from_probabilities_list, \
    get_frequencies_array_from_results
from qrem.functions_qrem.PyMaLi import GeneralTensorCalculator # old version: from PyMaLi import GeneralTensorCalculator
from qrem.functions_qrem import povmtools
from qrem.functions_qrem import functions_hyperplane_projections as fhp, ancillary_functions as anf

# from qrem.common.printer import qprint_array

# (PP) possible overlap with kronecker_product in math.py
# GTC stands for General Tensor Calculator.
def gtc_tensor_counting_function(arguments: list):
    result = 1

    for a in arguments:
        result = np.kron(result, a)

    return result


def gtc_matrix_product_counting_function(arguments: list):
    result = arguments[0]

    i = 1

    while i < len(arguments):
        result = result @ arguments[i]
        i += 1

    return result


#JT: The class below is used in reconstruction of POVMs

class QDTCalibrationSetup:
    """
        This class contains information required by DetectorTomographyFitter object to properly calculate
        maximum-likelihood POVM. This class shouldn't have any accessible methods and should only store and transfer
        data to the DetectorTomographyFitter class instances.
    """
    def __init__(self,
                 qubits_number: int,
                 probe_kets: List[np.array],
                 frequencies_array: np.ndarray,
                 states_labels:List[str]=None,
                 initial_ml_povm_guess = None):
        """
        Description:
            This is default constructor for QDTCalibrationSetup objects. It requires all necessary information, that is
            later used in the QDT process using DetectorTomographyFitter object.
        :param qubits_number: Number of qubits used in the circuits.
        :param probe_kets: Kets upon which circuits were build.
        :param frequencies_array: Results of circuits execution presented as frequencies.
        """
        self.qubits_number = qubits_number
        self.probe_kets = probe_kets

        # if probe_kets[0].shape[0]==2:
        #     self.probe_states = self.__get_probe_states(qubits_number, probe_kets)
        # else:

        if len(probe_kets[0].shape)>1:
            if probe_kets[0].shape[1]>1:
                self.probe_states = probe_kets
            else:

                self.probe_states = [ket@np.matrix.getH(ket) for ket in probe_kets]
        else:
            self.probe_states = [ket @ np.matrix.getH(ket) for ket in probe_kets]

        # from qrem.functions_qrem import ancillary_functions as anf
        # for pi in self.probe_states:
        #     qprint_array(pi)
        #
        # #
        # raise KeyboardInterrupt("jj")




        self.frequencies_array = frequencies_array

        self.states_labels = states_labels

        self.initial_ml_povm_guess = initial_ml_povm_guess


    @classmethod
    def from_qiskit_results(cls, results_list: List[Result], probe_kets: List[np.array]):
        """
        Description:
            This method_name generates Calibration setup objects directly from qiskit job results and probe kets used
            to generate circuits for these jobs_list. This method_name should be interpreted as sort of additional constructor
            for qiskit users.
        :param results_list: List of qiskit jobs_list results. In case of single job result it should still be a list.
        :param probe_kets: Prove kets (in form of list of np.arrays) used to generate calibration circuits.
        :return: Instance of QDT calibration setup from given job.
        """
        frequencies_array = get_frequencies_array_from_results(results_list)
        # This qubits_number calculation is a little elaborate, but necessary.
        circuits_number = sum(len(results.results) for results in results_list)
        qubits_number = int(log(circuits_number, len(probe_kets)))
        return cls(qubits_number, probe_kets, frequencies_array)

    @classmethod
    def from_frequencies_list(cls,
                              frequencies_list: List[Union[List[float], np.ndarray]],
                              probe_kets: List[np.array],
                              reverse_order: Optional[bool] = True
                              ):
        """
        Description:
            This method_name generates Calibration setup objects directly from qiskit job results and probe kets used
            to generate circuits for these jobs_list. This method_name should be interpreted as sort of additional constructor
            for qiskit users.
        :param frequencies_list: list of probabilities estimated in experiments

        :param probe_kets: Prove kets (in form of list of np.arrays) used to generate calibration circuits.
        :param reverse_order: specify whether probabiltiies lists should be reordered, which corresponds
                          to changing qubits' ordering
        :return: Instance of QDT calibration setup from given job.
        """
        number_of_qubits = int(np.log2(len(frequencies_list[0])))
        frequencies_array = get_frequencies_array_from_probabilities_list(frequencies_list=frequencies_list,
                                                                          reverse_order=reverse_order)

        return cls(number_of_qubits,
                   probe_kets,
                   frequencies_array)



    @staticmethod
    def __get_probe_states(qubits_number: int,
                           probe_kets: List[np.array]) -> List[np.ndarray]:
        """
        Description:
            This method_name generates probe states (density matrix) from results and kets
            passed to maximum likelihood POVM counting object.
        Parameters:
            :param qubits_number: Number of qubits used in the calibration experiments.
            :param probe_kets: Kets on which job circuits were based.
        Returns:
            List of probe states. These are supposed to have dimension equal to the size of Hilbert space, hence if one
            have used tensor products of single-qubit states, then one needs to give here those tensor products. Order
            needs to fit this of results.results.
        """

        probe_states = []

        for i in range(qubits_number):
            probe_states.append([get_density_matrix(ket) for ket in probe_kets])

        general_tensor_calculator = GeneralTensorCalculator(gtc_tensor_counting_function)

        return general_tensor_calculator.calculate_tensor_to_increasing_list(probe_states)


class QDTFitter:
    """
        This class is meant to resemble qiskit's state tomography and process tomography fitters and to calculate the
        maximum likelihood povm estimator describing a detector basing
        on QDT job results and used probe states.
    """

    def __init__(self, algorithm_convergence_threshold=1e-6):
        self.algorithmConvergenceThreshold = algorithm_convergence_threshold


    @staticmethod
    def __get_r_operator(m_m: np.ndarray, index_of_povm_effect: int, frequencies_array: np.ndarray,
                         probe_states: List[np.ndarray]) -> np.ndarray:
        """
        Description:
            This method_name calculates R operator as defined in Ref. [1].
        Parameters:
            :param m_m: Effect for which R operator is calculated.
            :param index_of_povm_effect: Index of povm effect for which R is calculated.
            :param frequencies_array: frequencies_array - array with size (m x number_of_qubits), where m means number of probe states,
            number_of_qubits means number of POSSIBLE outcomes.
            :param probe_states: arrray_to_print list of probe states density matrices.
        Returns:
            The R operator as described in Ref. [1].
        """

        number_of_probe_states = frequencies_array.shape[0]


        d = probe_states[0].shape[0]

        m_r = np.zeros((d, d), dtype=complex)

        for probe_state_index in range(number_of_probe_states):
            expectation_value_on_probe_state = np.trace(m_m @ probe_states[probe_state_index])

            if expectation_value_on_probe_state == 0j:
                continue

            x = frequencies_array[probe_state_index, index_of_povm_effect] / expectation_value_on_probe_state
            m_r += x * probe_states[probe_state_index]

        if np.linalg.norm(m_r) == 0:
            m_r = np.zeros((d, d), dtype=complex)

        return m_r

    @staticmethod
    def __get_lagrange_matrix(r_matrices: List[np.ndarray], povms: List[np.ndarray]) -> np.ndarray:
        """
        Description:
            Calculates Lagrange matrix used in Lagrange multipliers optimization method_name.
        Parameters:
            :param r_matrices: arrray_to_print list of R matrices described in a method_name generating them.
            :param povms: arrray_to_print list of effects for which Lagrange matrix will be calculated.
        Returns:
           Lagrange matrix for given parameters.
        """
        number_of_outcomes = len(povms)
        dimension = povms[0].shape[0]
        second_power_of_lagrange_matrix = np.zeros((dimension, dimension), dtype=complex)

        for j in range(number_of_outcomes):
            second_power_of_lagrange_matrix += r_matrices[j] @ povms[j] @ r_matrices[j]

        lagrange_matrix = sc.linalg.sqrtm(second_power_of_lagrange_matrix)

        return lagrange_matrix

    @staticmethod
    def __calculate_symmetric_m(m_lagrange_matrix: np.ndarray, m_r: np.ndarray, m_m: np.ndarray) -> np.ndarray:
        """
        Description:
            arrray_to_print method_name used for calculating symmetric m matrix.
        Parameters:
            :param m_m: arrray_to_print matrix of which symmetric method_name will be calculated.
            :param m_r: Previously calculated R operator.
            :param m_lagrange_matrix:
        Returns:
            Symmetric m matrix.
        """
        try:
            # Try to perform inversion of lagrange matrix.
            m_inversed_lagrange_matrix = np.linalg.inv(m_lagrange_matrix)
        except np.linalg.LinAlgError:
            # In some special cases it may fail. Provide identity matrix in that scenario.
            m_inversed_lagrange_matrix = np.eye(np.shape(m_lagrange_matrix)[0])

        symmetric_m = m_inversed_lagrange_matrix @ m_r @ m_m @ m_r @ m_inversed_lagrange_matrix

        return symmetric_m




    def _get_maximum_likelihood_povm_estimator(self,
                                               calibration_setup: QDTCalibrationSetup) -> List[np.ndarray]:
        """
        Description:
            Given results of Quantum Detector Tomography experiments and list of probe states, return the Maximum
            Likelihood estimation of POVM describing a detector. Uses recursive method_name from [1]. See also [2].
        Parameters:
            :param calibration_setup: QDTCalibrationSetup object that consists data upon which maximum likelihood POVM
            estimator should be calculated.
        Returns
            Maximum likelihood estimator of POVM describing a detector.
        """

        number_of_probe_states = calibration_setup.frequencies_array.shape[0]
        number_of_outcomes = calibration_setup.frequencies_array.shape[1]

        dimension = calibration_setup.probe_states[0].shape[0]

        from qrem.functions_qrem import ancillary_functions as anf

        # for pi in calibration_setup.probe_states:
        #
        #     qprint_array(pi)
        #
        # raise KeyboardInterrupt
        initial_povm = calibration_setup.initial_ml_povm_guess

        if initial_povm is None:

            povm = []


            for j in range(number_of_outcomes):
                povm.append(np.identity(dimension) / number_of_outcomes)
        else:
            povm = copy.deepcopy(initial_povm)

        # print(np.round(sum(povm),3))
        # # print(calibration_setup.frequencies_array.shape)
        # raise KeyError
        # Threshold is dynamic, thus another variable
        threshold = self.algorithmConvergenceThreshold

        i = 0
        current_difference = 1

        t0 = time.time()

        try:
            while abs(current_difference) >= threshold:
                i += 1

                if i % 50 == 0:
                    last_step_povm = copy.copy(povm)

                r_matrices = [self.__get_r_operator(povm[j], j,
                                                    calibration_setup.frequencies_array,
                                                    calibration_setup.probe_states)
                              for j in range(number_of_outcomes)]
                lagrange_matrix = self.__get_lagrange_matrix(r_matrices, povm)
                povm = [self.__calculate_symmetric_m(lagrange_matrix, r_matrices[j], povm[j])
                        for j in range(number_of_outcomes)]

                if i % 100 == 0:  # calculate the convergence tests only sometimes to make the code faster
                    current_difference = sum([np.linalg.norm(povm[k] - last_step_povm[k], ord=2)
                                              for k in range(number_of_outcomes)])




                elif (time.time()-t0)/60 > 30:  # make sure it does not take too long, sometimes convergence might not be so good
                    threshold = 1e-3

                elif (time.time()-t0)/60 > 10:  # make sure it does not take too long, sometimes convergence might not be so good
                    threshold = 1e-4

                if i%5000==0 and (time.time()-t0)>60:
                    print('iteration:',i,', with current difference: ',current_difference)

                    # for Mi in povm:
                    # qprint_array(povm[0])

        except(KeyboardInterrupt):
            pass


        return povm




    def _get_least_squares_povm_estimator_pauli(self,
                                                calibration_setup: QDTCalibrationSetup):


        number_of_probe_states = calibration_setup.frequencies_array.shape[0]
        number_of_outcomes = calibration_setup.frequencies_array.shape[1]
        dimension = calibration_setup.probe_states[0].shape[0]

        number_of_qubits = int(np.log2(dimension))

        norm = 1/3**number_of_qubits

        #First index - probe ket
        #Second index - outcome label
        frequencies_array = calibration_setup.frequencies_array
        probe_states = calibration_setup.probe_states


        states_labels = calibration_setup.states_labels

        povm_choi = np.zeros((dimension**2,dimension**2),dtype=complex)




        povm = [np.zeros((dimension,dimension),dtype=complex)
                for _ in range(number_of_outcomes)]
        #
        # from qrem.functions_qrem import ancillary_functions as anf
        # for pi in probe_states:
        #     qprint_array(pi)
        # raise KeyboardInterrupt("hejka")
        # print(number_of_outcomes,number_of_probe_states)
        for outcome_index in range(number_of_outcomes):
            for state_index in range(len(states_labels)):
                povm[outcome_index]+=probe_states[state_index]*frequencies_array[state_index,
                                                                                 outcome_index]

        povm = [Mi.T*norm for Mi in povm]

        return povm



    def __get_least_sqaures_povm_choi_estimator_pauli(self,
                                         calibration_setup):

        povm = self._get_least_squares_povm_estimator_pauli(calibration_setup=calibration_setup)
        return povmtools.get_choi_from_POVM(povm)

    def _get_projected_least_squares_povm_estimator(self,
                                                   calibration_setup):

        number_of_outcomes = calibration_setup.frequencies_array.shape[1]
        dimension = calibration_setup.probe_states[0].shape[0]

        number_of_qubits = int(np.log2(dimension))


        povm_LS = self._get_least_squares_povm_estimator_pauli(calibration_setup=calibration_setup)



        #JT a test is performed wheter POVM is a physical POVM, if this is true it is returned

        if povmtools.check_if_povm(povm=povm_LS):
            return povm_LS

        #JT: if this is not the case projection is performed

        else:

            #JT: here a Choi state of the corresponding POVM is performed

            choi_LS = povmtools.get_choi_from_POVM(POVM=povm_LS)

            #JT: Here Qutip is used, an object corresponding to POVM joi state is constructed

            test_choi = qutip.Qobj(choi_LS,superrep='choi',type='super')


            #JT: Now a test of complete positivity of the Choi state is performed

            if test_choi.iscp:

                #JT: If the state is completly positive no projection onto CP states is performed
                projected_choi_CP1 = choi_LS
            else:

                #JT: if the corresponding Choi state is not CP a projection is performed

                projected_choi_CP1 = fhp.new_proj_CP_threshold(rho=choi_LS,
                                                           free_trace=False,
                                                           thres_least_ev=True,
                                                           full_output=False
                                                           )
            #JT: The second projection (onto trace preserving - TP maps) is performed
            #JT: Again qutip is used

            projected_choi_CP_test = qutip.Qobj(projected_choi_CP1,superrep='choi',type='super')

            #JT: If the state is CPTP no projection is performed

            if projected_choi_CP_test.iscptp:
                projected_choi_CPTP = projected_choi_CP1

            #JT: If state is CP but not TP the final mixing with a CPTP state is performed

            else:
                projected_choi_CPTP = fhp.final_CPTP_by_mixing(fhp.proj_TP(projected_choi_CP1))



            #JT: When a proper Choi state is obtained, the resulting povm is constructed out of this Choi state

            povm = povmtools.get_POVM_from_choi(povm_choi=projected_choi_CPTP,
                                                number_of_outcomes=number_of_outcomes,
                                                dimension = int(2**number_of_qubits))
        return povm


    def get_POVM_estimator(self,
                           calibration_setup,
                           method='pls'):

        if method.lower() in ['maximal_likelihood', 'ml']:
            return self._get_maximum_likelihood_povm_estimator(calibration_setup=calibration_setup)
        elif method.lower() in ['least_squares', 'ls']:
            return self._get_least_squares_povm_estimator_pauli(calibration_setup=calibration_setup)
        elif method.lower() in ['projected_least_squares', 'pls']:
            return self._get_projected_least_squares_povm_estimator(calibration_setup=calibration_setup)




    # def get_least_squared_povm_estimator(self):







# TODO TR: This method_name may need to be revisited and possibly reduced into several smaller ones.
def join_povms(povms: List[List[np.ndarray]],
               qubit_indices_lists: List[List[int]],
               sort_outcomes = True) -> List[np.ndarray]:
    """
    Description:
        Generates a POVM from given list of POVMs and qubit indices.
    Parameter:
        :param povms: List of POVMs corresponding to qubits indices.
        :param qubit_indices_lists: Indices of qubits for which POVMs were calculated.
        :param sort_outcomes: indicates whether to sort effects of POVM in such that they correspond
                              to standard classical register -- WORKS ONLY FOR d-outcome measurements!
    Return:
        POVM describing whole detector.
    """
    qubits_num = sum([len(indices) for indices in qubit_indices_lists])
    raise KeyboardInterrupt
    swapped_povms = []

    for i in range(len(qubit_indices_lists)):
        indices_now = qubit_indices_lists[i]
        povm_now = povms[i]

        # extend povm to higher-dimensional space.by multiplying by complementing identity from right
        povm_now_extended = [np.kron(Mi, np.eye(2 ** (qubits_num - len(indices_now)))) for Mi in povm_now]

        # begin swapping
        swapped_povm = copy.copy(povm_now_extended)

        # go from back to ensure commuting
        for j in range(len(indices_now))[::-1]:
            index_qubit_now = indices_now[j]

            if index_qubit_now != j:
                # swap qubit from jth position to proper position
                swapped_povm = [permute_matrix(Mj, qubits_num, (j + 1, index_qubit_now + 1)) for Mj in swapped_povm]

        swapped_povms.append(swapped_povm)

    # With POVMs now represented by proper matrices (that is, parts corresponding to adequate qubits are now in
    # desired places), we want to join them.
    general_tensor_calculator = GeneralTensorCalculator(gtc_matrix_product_counting_function)
    povm = general_tensor_calculator.calculate_tensor_to_increasing_list(swapped_povms)

    if sort_outcomes:
        # We've obtained a POVM, but it is still ordered according to qubit indices. We want to undo that.
        indices_order = []
        for indices_list in qubit_indices_lists:
            indices_order = indices_order + indices_list

        new_classical_register = reorder_classical_register(indices_order)
        # print(new_classical_register,indices_order)

        sorted_povm = utils.sort_things(povm, new_classical_register)

        return sorted_povm
    else:
        return povm


import random
from qrem.functions_qrem import ancillary_functions as anf
from qrem.cn.simtools import auxiliary_merging_functions as m_f

from qrem.types import CNModelData as CNModelData 
from qrem.common import probability

import numpy as np

from typing import Tuple, Dict, List, Optional
import time

#TODO TUZJAN: Correct docstrings and place them inside functions 



#As for now neighbors functionality is not developed  

def create_random_noise_model(number_of_qubits:int,clusters_specification:List[List],neighbors:List[Tuple]=None) -> type[CNModelData]:
    '''
    Creates CNModelData with random noise. It used to generate data for tutorials. 
    Qubits are randomly assigned to clusters, number of clusters and their locality is specified by clusters_specification.
    Noise matrices assigned for clusters are random stochastic matrices. 

    

    Parameters
    ----------
    number_of_qubits - a number of qubits for which model is created 

    clusters_specification - a nested list containing description of clusters to be created:

            structure of the list [[element_1, element_2], ...]:
                element_1 - integer encoding locality of a cluster
                element_2 - number of clusters of given locality
            
            e.g. [[3,1],[2,2]] encodes 1 3 qubit cluster and two 2 qubit clusters

    neighbors - an optional parameter for neighbors specification, by default set to None 

    Returns
    -------
        noise_model: CNModelData object
             CNModelData object storing noise model data

    '''
    #[1] Create empty holder for model data CNModelData
    noise_model = CNModelData(number_of_qubits=number_of_qubits)

    #[2] Now qubits are randomly assigned to clusters of a provided structure
    clusters_tuple=m_f.divide_qubits_in_the_clusters([i for i in range(number_of_qubits)],clusters_specification=clusters_specification)
    noise_model.set_clusters_tuple(clusters_tuple=clusters_tuple)

    
    #[3] Random stochastic matrix is generated per each cluster, data is stored in a dictionary
    noise_model_dictionary ={}    

    if neighbors==None:
        #[3.1] TODO: now we don't suport neighbours, so it should be always true
        for element in clusters_tuple:
            noise_model_dictionary[element] = probability.random_stochastic_matrix(2**len(element))
    noise_model.set_noise_matrices_dictionary(noise_matrices_dictionary=noise_model_dictionary)

    #[4] We should be set with simulated CNModelData
    return noise_model


def create_custom_noise_model (number_of_qubits:int, noise_matrices_dictionary:dict) -> type[CNModelData]:
    
    '''
    Creates CNModelData with data provided by the user. It used to debug mitigation routines. 
     

    

    Parameters
    ----------
    number_of_qubits: integer 
        A number of qubits for which model is created
    
    noise_matrices_dictionary: dictionary
        Dictionary specifying clusters and noise matrices. Key correspond to tupels of qubit indices belonging to the same cluster, values to corresponding nosie matrices.
        E.g . (0,2) : 4 by 4 stochastic matrix    


   
    Returns
    -------
        noise_model: CNModelData object
             CNModelData object storing noise model data
    '''
    
    
    #[1] Create empty holder for model data CNModelData
    noise_model = CNModelData(number_of_qubits=number_of_qubits)
    
    #[2] Object is filled with data provided by 
    noise_model.set_noise_model(noise_matrices_dictionary=noise_matrices_dictionary)

    return noise_model






#This function takes a dictionary consisting of DOT experiment setting (string describing unitaries) and result and transforms it accordingly to a given noise model
#The noise model is specified accordingly to QREM convection, dictionary where keys are tuples of integers encoding qubits belonging to a cluster and associated array of noise matrices on clusters (key averaged indicates absence of neighborhoods )


def simulate_noise_results_dictionary(results_dictionary: Dict[str, Dict[str, int]],
                            noise_model:type[CNModelData],
                            neighborhoods_tuples: Optional[Dict[Tuple[int], List[int]]]={}
                            ) -> Dict[str ,Dict]:
    '''
    Function simulates readout noise according to a noise model specified CNModelData on experimental data.
    It is used in tutorials. Function loops over provided (noiseless) results dictionary and transforms the
    counts according to a product probability distribution specified by a CN noise model.  

    Parameters
    ----------

    results_dictionary: dictionary 
        dictionary with results of an experiment

    noise_model: CNModelData object
        CNModelData object storing noise model data
    
    Returns
    ----------
    noisy_results_total
            dictionary with results of a an experiment with added readout noise


    '''
    noisy_results_dictionary={}

    for input_state, noiseless_results in results_dictionary.items():
        
        noisy_results_list = []
        cluster_samples_dict = {}
        for single_result, number_of_samples in noiseless_results.items():
            cluster_samples_dict[single_result]={}
            t0=time.time()
            for cluster_qubits, noise_matrix_on_cluster in (noise_model.noise_matrices).items():

                state_on_cluster = ''.join([single_result[x] for x in cluster_qubits])
                
                if neighborhoods_tuples == {}:
                    noise_matrix = noise_matrix_on_cluster
                else:
                    noise_matrix = noise_matrix_on_cluster['averaged']

                prob_dist = noise_matrix[:,int(state_on_cluster,2)]

                cluster_dist = np.random.multinomial(n=number_of_samples, pvals=prob_dist)
                cluster_content = []
                for i in range(len(cluster_dist)):
                    for j in range(cluster_dist[i]):
                        cluster_content.append(i)

                cluster_sample = np.random.permutation(cluster_content)
                cluster_samples_dict[single_result][cluster_qubits]=cluster_sample
            

            for m in range(number_of_samples):
                
                state_string = list(single_result)

                for cluster_qubits in (noise_model.noise_matrices).keys():
                    k = len(cluster_qubits)
                    substring = f'{cluster_samples_dict[single_result][cluster_qubits][m]:0{k}b}'
                    
                    for i in range(k):
                        state_string[cluster_qubits[i]]=substring[i]
                    
                
                noisy_results_list.append("".join(state_string))
                
            
            
        
        noisy_results_dictionary[input_state]={}
        single_input_dict = noisy_results_dictionary[input_state]

        for result in noisy_results_list:
            if result in single_input_dict.keys():
                single_input_dict[result]+=1
            else:
                single_input_dict[result]=1
        
    
    return noisy_results_dictionary




def simulate_noise_results_dictionary_old(results_dictionary: Dict[str, Dict[str, int]],
                            noise_model:type[CNModelData],
                            neighborhoods_tuples: Optional[Dict[Tuple[int], List[int]]]={}
                            ) -> Dict[str ,Dict]:
    
    '''
    DEPRECATED: This function is replaced by the optimized version simulate_noise_results_dictionary() 
    '''
    
    


    #_noisy_results_total - a dictionary that holds results of noisy experiments, at the beginning empty

    noisy_results_total={}
    # an outer loop over settings and results from results_dictionary, a dictionary storing results of ideal (noiseless) experiment
    for settings, results_for_settings in results_dictionary.items():

        #an array for new noise samples, noise is added sample by sample, the qubit ordering according to the qubit cluster list
        
        new_results = np.array([ 0 for _ in range(2**len(list(results_for_settings.keys())[0]))])

        # for each input state a probability distribution encoding noise is created
        t0 = time.time()
        for input_state, number_of_samples in results_for_settings.items():

            

            # this loop goes over clusters and respective noise matrices, as for now we support only the simplified model without neighbors

            prob_dist_now =[]
            marginal_samples =[]
            first=True
            dim_total=1
            print(f"    {input_state},{number_of_samples}")
            for cluster_qubits, noise_matrices_on_cluster in (noise_model.noise_matrices).items() :
                
                

                dim_now=2**len(cluster_qubits)

               

            
                
                #a state of qubits belonging to the current cluster is established

                input_on_cluster = ''.join([input_state[x] for x in cluster_qubits])

                #the code below is prepared for clusters extension, which is not implemented now
                #TODO: FINISH this
                
                #cluster noise matrix is specified
                if neighborhoods_tuples == {}:
                    noise_matrix = noise_matrices_on_cluster
                else:
                    noise_matrix = noise_matrices_on_cluster['averaged']

                        

                # a transfer probability is chosen for a given state of the results

                prob_dist_now=noise_matrix[:, int(input_on_cluster, 2)]

                
         

                #samples are created according to respective product probabity distributions
                # the if statement is needed to handle the case when a first sample is crated (there are no other samples to join)
                # the general idea of joining different samples is the following
                # 1) a list of counts is crated according to a probability distribution e.g. for two qubits it is of a form [4,8,1,7] 
                # for an experiment with 20 shots
                # 2) The list is transformed to a randomized result list e.g [2,0,1,3,...]
                # 3) This list is joined with a previously generated measurement statistic
                                  
                if first:
                    marginal_samples = np.random.multinomial(n=number_of_samples,pvals=prob_dist_now)
                    marginal_samples=m_f.transform_marginal_statistics_to_list_of_results(marginal_samples)
                    
                    dim_total=dim_total*dim_now
                    
                    first=False
                else:

                      

                    marginal_samples_temp=m_f.transform_marginal_statistics_to_list_of_results(np.random.multinomial(n=number_of_samples,pvals=prob_dist_now))

                   
                    marginal_results=m_f.merge_samples_list(marginal_samples,dim_total,marginal_samples_temp,dim_now)

                    marginal_samples = m_f.transform_marginal_statistics_to_list_of_results(marginal_results)

                    dim_total=dim_total*dim_now
                
            #When the loop over clusters is finished new noisy result is updated 

            
            new_results = new_results + marginal_results
        t1=time.time()
        print(f"random sampling time {t1-t0} seconds")




           



            

        #In the above, the orders of qubits was that of clusters. Here the initial order of qubits is restored.

        noisy_results={}

        # this is a loop over all possible bitstrings (results) of qubits
        #the order of qubits is changed
        #key on clusters is a string corresponding to results

        key = ['0' for _ in  range(len(input_state))]

        for i in range(len(new_results)):

            #the bit string corresponding to results is created

            key_on_clusters = "{0:b}".format(2**len(input_state)+i)
            key_on_clusters = key_on_clusters[1:]

            #counter is a variable that goes over physical qubits

            counter=0

            #the first for loop is over different cluster assigment

            for qubits_in_clusters in noise_model.clusters_tuple:

                #the second for loop is over qubits belonging to that cluster

                for qubit in qubits_in_clusters:
                    #order of qubits is corrected: in the noisy results qubit on counter place, in put on the place according to the qubit ordering
                    key[qubit] = key_on_clusters[counter]
                    counter+=1

            #now it is checked whether a given result appears in the noise results, if yes, it is assigned to a dictionary

            if new_results[i] != 0:

                noisy_results.update({''.join(key):new_results[i]})

        noisy_results_total.update({settings:noisy_results})
        t2=time.time()
        print(f"formatting time: {t2-t1} seconds")
    return noisy_results_total



#possibly faster version, test needed
#KKM: tried it ad hoc, it has some bugs
def simulate_noise_results_dictionary__new_version(results_dictionary: Dict[str, Dict[str, int]],
                            noise_model:type[CNModelData],
                            neighborhoods_tuples: Optional[Dict[Tuple[int], List[int]]]={}
                            ) -> Dict[str ,Dict]:
    
    '''
    DEPRECATED: This function is replaced by the optimized version simulate_noise_results_dictionary() 

    Function simulate readout noise according to a noise model specified CNModelData on experimental data.
    It is used in tutorials. Function loops over provided (noiseless) results dictionary and transforms the
    counts according to a product probability distribution specified by a CN noise model.  

    Parameters
    ----------

    results_dictionary: dictionary 
        dictionary with results of an experiment

    noise_model: CNModelData object
        CNModelData object storing noise model data
    
    Returns
    ----------
    noisy_results_total
            dictionary with results of a an experiment with added readout noise


    '''
    
    


    #_noisy_results_total - a dictionary that holds results of noisy experiments, at the beginning empty

    noisy_results_total={}
    # an outer loop over settings and results from results_dictionary, a dictionary storing results of ideal (noiseless) experiment
    
    for settings, results_for_settings in results_dictionary.items():
        #an array for new noise samples, noise is added sample by sample, the qubit ordering according to the qubit cluster list
        
        new_results = np.array([ 0 for _ in range(2**len(list(results_for_settings.keys())[0]))])

        # for each input state a probability distribution encoding noise is created
        t0 = time.time()
        for input_state, number_of_samples in results_for_settings.items():

            

            # this loop goes over clusters and respective noise matrices, as for now we support only the simplified model without neighbors

            prob_dist_now =[]
            marginal_samples =[]
            first=True
            dim_total=1
            for cluster_qubits, noise_matrices_on_cluster in (noise_model.noise_matrices).items() :
                
                

                dim_now=2**len(cluster_qubits)

               

            
                
                #a state of qubits belonging to the current cluster is established

                input_on_cluster = ''.join([input_state[x] for x in cluster_qubits])

                #the code below is prepared for clusters extension, which is not implemented now
                #TODO: FINISH this
                
                #cluster noise matrix is specified
                if neighborhoods_tuples == {}:
                    noise_matrix = noise_matrices_on_cluster
                else:
                    noise_matrix = noise_matrices_on_cluster['averaged']

                        

                # a transfer probability is chosen for a given state of the results

                prob_dist_now=noise_matrix[:, int(input_on_cluster, 2)]

                
         

                #samples are created according to respective product probabity distributions
                # the if statement is needed to handle the case when a first sample is crated (there are no other samples to join)
                # the general idea of joining different samples is the following
                # 1) a list of counts is crated according to a probability distribution e.g. for two qubits it is of a form [4,8,1,7] 
                # for an experiment with 20 shots
                # 2) The list is transformed to a randomized result list e.g [2,0,1,3,...]
                # 3) This list is joined with a previously generated measurement statistic
                                  
                if first:
                    marginal_samples = np.random.multinomial(n=number_of_samples,pvals=prob_dist_now)
                    #marginal_samples=m_f.transform_marginal_statistics_to_list_of_results(marginal_samples)
                    
                    #dim_total=dim_total*dim_now
                    
                    first=False
                else:

                      
                    marginal_samples_temp=np.random.multinomial(n=number_of_samples,pvals=prob_dist_now)    
                    #marginal_samples_temp=m_f.transform_marginal_statistics_to_list_of_results(np.random.multinomial(n=number_of_samples,pvals=prob_dist_now))

                   
                    #marginal_results=m_f.merge_samples_list(marginal_samples,dim_total,marginal_samples_temp,dim_now)

                    marginal_samples = m_f.merge_results_list(marginal_samples,marginal_samples_temp)

                    #dim_total=dim_total*dim_now

            #When the loop over clusters is finished new noisy result is updated 

            
            new_results = new_results + marginal_samples
        t1=time.time()
        print(f"random sampling time {t1-t0} seconds")
        




           



            

        #In the above, the orders of qubits was that of clusters. Here the initial order of qubits is restored.

        noisy_results={}

        # this is a loop over all possible bitstrings (results) of qubits
        #the order of qubits is changed
        #key on clusters is a string corresponding to results

        key = ['0' for _ in  range(len(input_state))]

        for i in range(len(new_results)):

            #the bit string corresponding to results is created

            key_on_clusters = "{0:b}".format(2**len(input_state)+i)
            key_on_clusters = key_on_clusters[1:]

            #counter is a variable that goes over physical qubits

            counter=0

            #the first for loop is over different cluster assigment

            for qubits_in_clusters in noise_model.clusters_tuple:

                #the second for loop is over qubits belonging to that cluster

                for qubit in qubits_in_clusters:
                    #order of qubits is corrected: in the noisy results qubit on counter place, in put on the place according to the qubit ordering
                    key[qubit] = key_on_clusters[counter]
                    counter+=1

            #now it is checked whether a given result appears in the noise results, if yes, it is assigned to a dictionary

            if new_results[i] != 0:

                noisy_results.update({''.join(key):new_results[i]})

        noisy_results_total.update({settings:noisy_results})
        t2=time.time()
        print(f"formatting time: {t2-t1} seconds")
    return noisy_results_total





            




import copy
from collections import defaultdict
from typing import Optional, Dict, List, Union, Tuple

import numpy as np
from tqdm import tqdm


from qrem.functions_qrem import ancillary_functions as anf
from qrem.noise_characterization.base_classes.marginals_analyzer_base import MarginalsAnalyzerBase

import qrem.common.math as qrem_math
from qrem.common import utils
from qrem.common.printer import qprint, qprint_array


#(PP): DELETE, should be removed from package everywhere it is used (no human input needed rule)
def query_yes_no(question):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    yes_answers = {'yes',
                   'y',
                   'ye',
                   'tak',
                   'sure',
                   'of course',
                   'Yes',
                   'yeah'}
    no_answers = {'no',
                  'n',
                  'nope',
                  'nah',
                  'nie',
                  'noo',
                  'nooo',
                  'noooo',
                  'No'}

    choice = 0
    print(question + ' [y/n] ')
    choice = input().lower()
    if choice in yes_answers:
        return True
    elif choice in no_answers:
        return False
    else:
        qprint('Please:', "respond with 'yes' or 'no'")
        return query_yes_no(question)


class DOTMarginalsAnalyzer(MarginalsAnalyzerBase):
    """
        Class that handles results of Diagonal Detector Tomography.
        Main functionalities allow to calculate noise matrices on subsets_list of qubits.
        This includes averaged noise matrices, i_index.e., averaged over states off all other qubits,
        as well as state-dependent, i_index.e., conditioned on the particular
        input classical state of some other qubits.

        In this class, and all its children, we use the following convention
        for storing marginal noise matrices:

        :param noise_matrices_dictionary: nested dictionary with following structure:

        noise_matrices_dictionary[qubits_subset_string]['averaged']
        = average noise matrix on qubits subset
        and
        noise_matrices_dictionary[qubits_subset_string][other_qubits_subset_string][input_state_bitstring]
        = noise matrix on qubits subset depending on input state of other qubits.

        where:
        - qubits_subset_string - is string labeling qubits subset (e.g., 'q1q2q15...')
        - other_qubits_subset_string - string labeling other subset
        - input_state_bitstring - bitstring labeling
                                                   input state of qubits in other_qubits_subset_string

    """

    def __init__(self,
                 results_dictionary_ddot: Dict[str, Dict[str, int]],
                 bitstrings_right_to_left: Optional[bool] = False,
                 marginals_dictionary: Optional[Dict[str, Dict[Tuple[int], np.ndarray]]] = None,
                 noise_matrices_dictionary: Optional[
                     Dict[Tuple[int], Dict[Union[str,Tuple[int]], Dict[str, np.ndarray]]]] = None
                 ) -> None:

        """
        :param results_dictionary_ddot: see description of MarginalsAnalyzerBase.
        Here we use classical input states (bitstrings) of qubits as LABELS for experiments.
        :param bitstrings_right_to_left: specify whether bitstrings
                                 should be read from right to left (when interpreting qubit labels)
        :param marginals_dictionary: see description of MarginalsAnalyzerBase.
        :param noise_matrices_dictionary: nested dictionary with following structure:
        """

        super().__init__(results_dictionary=results_dictionary_ddot,
                         bitstrings_right_to_left=bitstrings_right_to_left,
                         marginals_dictionary=marginals_dictionary
                         )


        self.__fill_missing_columns = False



        if noise_matrices_dictionary is None:
            noise_matrices_dictionary = {}
            # TODO FBM: Make sure whether this helps in anything
            # TODO FBM: (because we anyway perform checks in the functions later)
            if marginals_dictionary is not None:
                for experiment_key, dictionary_of_marginals in marginals_dictionary.items():
                    for marginal_key in dictionary_of_marginals.keys():
                        if marginal_key not in noise_matrices_dictionary.keys():
                            noise_matrices_dictionary[marginal_key] = {}

        self._noise_matrices_dictionary = noise_matrices_dictionary

    @property
    def noise_matrices_dictionary(self) -> Dict[
        Tuple[int], Dict[Union[str, Tuple[int]], Dict[str, np.ndarray]]]:
        return self._noise_matrices_dictionary

    @noise_matrices_dictionary.setter
    def noise_matrices_dictionary(self,
                                  noise_matrices_dictionary: Dict[Tuple[int],
                                                                  Dict[Union[str, Tuple[int]],
                                                                       Dict[
                                                                           str, np.ndarray]]] = None) -> None:

        self._noise_matrices_dictionary = noise_matrices_dictionary

    @staticmethod
    def get_noise_matrix_from_counts_dict(
            results_dictionary: Union[Dict[str, np.ndarray], defaultdict]) -> np.ndarray:
        """Return noise matrix from counts dictionary.
        Assuming that the results are given only for qubits of interest.
        :param results_dictionary: dictionary with experiments of the form:

        results_dictionary[input_state_bitstring] = probability_distribution

        where:
        - input_state_bitstring is bitstring denoting classical input state
        - probability_distribution - estimated vector of probabilities for that input state

        :return: noise_matrix: the array representing noise on qubits
        on which the experiments were performed
        """

        number_of_qubits = len(list(results_dictionary.keys())[0])
        # print('hej',number_of_qubits)
        noise_matrix = np.zeros((2 ** number_of_qubits, 2 ** number_of_qubits))
        for input_state, probability_vector in results_dictionary.items():
            numbers_input = [int(x) for x in list(input_state)]

            #TODO FBM: this should never happen, right? Control it
            if np.any(np.array(numbers_input) > 1):
                continue


            try:
                noise_matrix[:, int(input_state, 2)] = probability_vector[:]
            except(ValueError):
                noise_matrix[:, int(input_state, 2)] = probability_vector[:, 0]



        return noise_matrix

    @staticmethod
    def average_noise_matrices_over_some_qubits(matrices_cluster: Dict[str, np.ndarray],
                                                all_neighbors: List[int],
                                                qubits_to_be_left: List[int]) -> Dict[str, np.ndarray]:
        """
          Given dictionary of noise matrices, average them over some qubits.

         :param matrices_cluster: dictionary for which KEY is classical INPUT state of neighbors,
                                  and VALUE is stochastic_matrix noise matrix
         :param all_neighbors: list of neighbors of given cluster
         :param qubits_to_be_left: qubits which we are interested in and we do not average over them

         :return: dictionary of noise matrices
                    depending on the state of neighbors MINUS qubits_to_be_averaged_over

         """

        if all_neighbors is None or len(all_neighbors) == 0:
            return {'averaged': matrices_cluster['averaged']}

        reversed_enumerated = utils.map_index_to_order(all_neighbors)
        averaging_normalization = int(2 ** (len(all_neighbors) - len(qubits_to_be_left)))

        states_after_averaging = anf.get_classical_register_bitstrings(range(len(qubits_to_be_left)),
                                                                       len(qubits_to_be_left), False)
        averaged_dimension = list(matrices_cluster.values())[0].shape[0]

        averaged_matrices_cluster = {
            state: np.zeros((averaged_dimension, averaged_dimension), dtype=float) for state in
            states_after_averaging}

        qubits_to_be_averaged_over = list(set(all_neighbors).difference(set(qubits_to_be_left)))
        qubits_to_be_averaged_over_mapped = [reversed_enumerated[q_index] for q_index in
                                             qubits_to_be_averaged_over]

        for neighbors_state, conditional_noise_matrix in matrices_cluster.items():
            list_string_neighbors = list(copy.deepcopy(neighbors_state))

            list_string_neighbors_to_be_left = list(np.delete(list_string_neighbors,
                                                              qubits_to_be_averaged_over_mapped))

            string_neighbors = ''.join(list_string_neighbors_to_be_left)

            averaged_matrices_cluster[string_neighbors] += conditional_noise_matrix / averaging_normalization

        return averaged_matrices_cluster

    def _compute_noise_matrix_averaged(self,
                                       subset: Tuple[int]) -> np.ndarray:
        """Noise matrix for subset of qubits, averaged over all other qubits

            :param subset: subset of qubits we are interested in

           By default takes data from self._marginals_dictionary. If data is not present, then it
           calculates marginals_dictionary for given subset
           and updates the class's property self.marginals_dictionary
        """

        # TODO FBM: Perhaps add possibility of using existing marginals_dictionary for bigger subset that includes target subset
        #

        # subset_key = 'q' + 'q'.join([str(s) for s in subset])
        subset_key = tuple(subset)



        marginal_dict_now = self.get_averaged_marginal_for_subset(subset)

        # print('hej')
        # print(marginal_dict_now)
        noise_matrix_averaged = self.get_noise_matrix_from_counts_dict(marginal_dict_now)


        # fill_missing = False

        if not qrem_math.is_matrix_stochastic(noise_matrix_averaged):
            # qprint_array(noise_matrix_averaged)
            message_now = f"\nNoise matrix not stochastic for subset: {subset}.\n" \
                          f"This most likely means that DDOT collection was not complete " \
                          f"for locality {len(subset)} and some " \
                          f"states were not implemented."
            qprint(message_now,'','red')
            qprint("\nThat matrix looks like this:")
            qprint_array(noise_matrix_averaged)

            print('\n')
            if self.__fill_missing_columns:
                qprint("\nAdding missing columns.",'','red')
                for column_index in range(noise_matrix_averaged.shape[0]):
                    column_now = noise_matrix_averaged[:, column_index]

                    if np.all(column_now == 0):
                        noise_matrix_averaged[column_index, column_index] = 1.0

                qprint("\nNow the matrix looks like this:")
                qprint_array(noise_matrix_averaged)

            elif query_yes_no("Do you wish to assume that missing columns are ideal?\n"
                                "Otherwise, we will raise ValueError."):

                for column_index in range(noise_matrix_averaged.shape[0]):
                    column_now = noise_matrix_averaged[:,column_index]
                    
                    if np.all(column_now==0):
                        noise_matrix_averaged[column_index,column_index] = 1.0

                qprint("\nNow the matrix looks like this:")
                qprint_array(noise_matrix_averaged)


                if query_yes_no("Do you wish to fill all missing columns in the future?\n"
                                    "Otherwise, we will ask each time where there's need."):
                    self.__fill_missing_columns = True
                    qprint("OK, from now on we will assume missing columns are ideal.\n")
                else:
                    qprint("OK, we will ask again if needed.")

            else:
                raise ValueError()

        if subset_key in self._noise_matrices_dictionary.keys():
            self._noise_matrices_dictionary[subset_key]['averaged'] = noise_matrix_averaged
        else:
            self._noise_matrices_dictionary[subset_key] = {'averaged': noise_matrix_averaged}

        return noise_matrix_averaged

    def get_noise_matrix_averaged(self,
                                  subset: Tuple[int]) -> np.ndarray:
        """
            Like self._compute_noise_matrix_averaged but if matrix is already in class' property,
            does not calculate it again.

            :param subset: subset of qubits we are interested in
        """
        # subset_key = 'q' + 'q'.join([str(s) for s in subset])
        subset_key = tuple(subset)

        # print('hej',subset)
        # print(self._noise_matrices_dictionary)
        if subset_key in self._noise_matrices_dictionary.keys():
            if 'averaged' in self._noise_matrices_dictionary[subset_key].keys():
                return self._noise_matrices_dictionary[subset_key]['averaged']
            else:
                return self._compute_noise_matrix_averaged(subset)
        else:
            return self._compute_noise_matrix_averaged(subset)

    #(PP) this function uses self.get_qubits_keystring which I think was NEVER DEFINED? Also scan of repo did not return any results...
    #BUG WARNING
    def _compute_noise_matrix_dependent(self,
                                        qubits_of_interest: Tuple[int],
                                        neighbors_of_interest: Union[Tuple[int], None]) \
            -> Dict[str, np.ndarray]:
        """Return lower-dimensional effective noise matrices acting on qubits_of_interest"
                    conditioned on input states of neighbors_of_interest
            :param qubits_of_interest: labels of qubits in marginal  we are interested in
            :param neighbors_of_interest: labels of qubits that affect noise matrix on qubits_of_interest

            :return conditional_noise_matrices_dictionary: dictionary with structure

            conditional_noise_matrices_dictionary['averaged'] =
            noise matrix on qubits_of_interest averaged over input states of other qubits

            and

            conditional_noise_matrices_dictionary[input_state_neighbors_bitstring] =
            noise matrix on qubits_of_interest conditioned on input state of neighbors being
            input_state_neighbors_bitstring

        """

        # If there are no all_neighbors,
        # then this corresponds to averaging over all qubits except qubits_of_interest
        if len(neighbors_of_interest) == 0 or neighbors_of_interest is None:
            cluster_tuple_now = anf.get_qubits_keystring(qubits_of_interest)
            cluster_tuple_now = tuple(qubits_of_interest)
            if 'averaged' in self._noise_matrices_dictionary[cluster_tuple_now].keys():
                return {'averaged': self._noise_matrices_dictionary[cluster_tuple_now]['averaged']}
            else:
                noise_matrix = self.get_noise_matrix_averaged(qubits_of_interest)
                return {'averaged': noise_matrix}

        # check if there is no collision between qubits_of_interest and neighbors_of_interest
        # (if there is, then the method_name won't be consistent)
        if len(utils.lists_intersection(qubits_of_interest, neighbors_of_interest)) != 0:
            print(qubits_of_interest, neighbors_of_interest)
            raise ValueError('Qubits of interest and neighbors overlap')

        # first, get averaged noise matrix on qubits of interest and all_neighbors of interest
        # TODO FBM: make sure that qubit indices are correct (I think they are)
        qubits_of_interest = list(qubits_of_interest)
        neighbors_of_interest = list(neighbors_of_interest)
        all_qubits = sorted(qubits_of_interest + neighbors_of_interest)
        all_qubits_enumerated = utils.map_index_to_order(all_qubits)

        # we will get noise matrix on all of the qubits first, and then we will process it to get
        # conditional marginal noise matrices on qubits_of_interest
        big_lambda = self.get_noise_matrix_averaged(tuple(all_qubits))

        total_number_of_qubits = int(np.log2(big_lambda.shape[0]))
        total_dimension = int(2 ** total_number_of_qubits)
        number_of_qubits_of_interest = len(qubits_of_interest)
        number_of_neighbors = len(neighbors_of_interest)

        # Normalization when averaging over states of non-neighbours (each with the same probability)
        normalization = 2 ** (
                total_number_of_qubits - number_of_neighbors - number_of_qubits_of_interest)

        # classical register on all qubits
        classical_register_all_qubits = ["{0:b}".format(i).zfill(total_number_of_qubits) for i in
                                         range(total_dimension)]

        # classical register on neighbours
        classical_register_neighbours = ["{0:b}".format(i).zfill(number_of_neighbors) for i in
                                         range(2 ** number_of_neighbors)]

        # create dictionary of the marginal states of qubits_of_interest and neighbors_of_interest
        # for the whole register (this function is storing data which could also be calculated in situ
        # in the loops later, but this is faster)
        indices_dictionary_small = {}
        for neighbors_state_bitstring in classical_register_all_qubits:
            small_string = ''.join([list(neighbors_state_bitstring)[all_qubits_enumerated[b]] for b in
                                    qubits_of_interest])
            neighbours_tuple_now = ''.join(
                [list(neighbors_state_bitstring)[all_qubits_enumerated[b]] for b in
                 neighbors_of_interest])
            # first place in list is label for state of qubits_of_interest
            # and second for neighbors_of_interest
            indices_dictionary_small[neighbors_state_bitstring] = [small_string, neighbours_tuple_now]

        # initiate dictionary for which KEY is input state of all_neighbors
        # and VALUE will the the corresponding noise matrix on qubits_of_interest
        conditional_noise_matrices = {
            s: np.zeros((2 ** number_of_qubits_of_interest, 2 ** number_of_qubits_of_interest)) for s
            in
            classical_register_neighbours}

        # go through all classical states
        for measured_state_integer in range(total_dimension):
            for input_state_integer in range(total_dimension):
                lambda_element = big_lambda[measured_state_integer, input_state_integer]

                # input state of all qubits in binary format
                input_state_bitstring = classical_register_all_qubits[input_state_integer]
                # measured state of all qubits in binary format
                measured_state_bitstring = classical_register_all_qubits[measured_state_integer]

                # input state of qubits_of_interest in binary format
                input_state_small = indices_dictionary_small[input_state_bitstring][0]
                # measured state of qubits_of_interest in binary format
                measured_state_small = indices_dictionary_small[measured_state_bitstring][0]

                # input state of neighbors_of_interest in binary format
                input_state_neighbours = indices_dictionary_small[input_state_bitstring][1]

                # element of small lambda labeled by (measured state | input state),
                # and the lambda itself is labeled by input state of all_neighbors
                conditional_noise_matrices[input_state_neighbours][
                    int(measured_state_small, 2), int(input_state_small, 2)] += lambda_element

        # normalize matrices
        for neighbors_state_bitstring in classical_register_neighbours:
            conditional_noise_matrices[neighbors_state_bitstring] /= normalization

        # conditional_noise_matrices['all_neighbors'] = neighbors_of_interest

        # cluster_string = 'q' + 'q'.join(str(s) for s in qubits_of_interest)
        # neighbours_string = 'q' + 'q'.join(str(s) for s in neighbors_of_interest)

        cluster_tuple_now = tuple(qubits_of_interest)
        neighbours_tuple_now = tuple(neighbors_of_interest)


        if cluster_tuple_now not in self._noise_matrices_dictionary.keys():
            # If there is no entry for our cluster in the dictionary, we create it and add
            # averaged noise matrix
            averaged_noise_matrix = np.zeros(
                (2 ** number_of_qubits_of_interest, 2 ** number_of_qubits_of_interest))
            for neighbors_state_bitstring in conditional_noise_matrices.keys():
                averaged_noise_matrix += conditional_noise_matrices[neighbors_state_bitstring]
            averaged_noise_matrix /= 2 ** number_of_qubits_of_interest
            self._noise_matrices_dictionary[cluster_tuple_now] = {'averaged': averaged_noise_matrix}

        self._noise_matrices_dictionary[cluster_tuple_now][neighbours_tuple_now] = conditional_noise_matrices

        return self._noise_matrices_dictionary[cluster_tuple_now][neighbours_tuple_now]


    def get_noise_matrix_dependent(self,
                                   qubits_of_interest: Tuple[int],
                                   neighbors_of_interest: Tuple[int]) -> dict:
        """Description:
        like self._compute_noise_matrix_dependent
        but checks whether matrices were already calculated to prevent multiple computations of the
        same matrices

        :param qubits_of_interest: labels of qubits in marginal  we are interested in
        :param neighbors_of_interest: labels of qubits that affect noise matrix on qubits_of_interest

        :return conditional_noise_matrices_dictionary:

        """

        # cluster_key = anf.get_qubits_keystring(qubits_of_interest)

        cluster_key = tuple(qubits_of_interest)
        if cluster_key not in self._noise_matrices_dictionary.keys():
            self.compute_subset_noise_matrices_averaged([qubits_of_interest])
        # print('hey')

        if len(neighbors_of_interest) == 0 or neighbors_of_interest is None:
            neighbors_key = 'averaged'

            if neighbors_key in self._noise_matrices_dictionary[cluster_key]:
                # if not qrem_math.is_matrix_stochastic(self._noise_matrices_dictionary[cluster_key]['averaged']):
                #     qprint('Bug is here')
                #     print(cluster_key, neighbors_key)
                #     # TODO FBM: SOMETHING IS BROKEN
                #     self._noise_matrices_dictionary[cluster_key][
                #         'averaged'] = self._compute_noise_matrix_averaged(qubits_of_interest)
                #     if not qrem_math.is_matrix_stochastic(self._noise_matrices_dictionary[cluster_key]['averaged']):
                #         qprint('And I cant fix it')

                # qprint_array(self._noise_matrices_dictionary[cluster_key]['averaged'])
                return {'averaged': self._noise_matrices_dictionary[cluster_key]['averaged']}
            else:

                #note that this is already a dict anyway
                averaged_matrix_dictionary = self._compute_noise_matrix_dependent(qubits_of_interest,
                                                                       neighbors_of_interest)

                if not qrem_math.is_matrix_stochastic(averaged_matrix_dictionary['averaged']):
                    raise ValueError(f"Noise matrix not stochastic for qubits: "
                                     f"cluster = {cluster_key}, neighbors = {neighbors_of_interest}.")

                return averaged_matrix_dictionary

        else:
            # neighbors_key = 'q' + 'q'.join([str(s) for s in neighbors_of_interest])

            neighbors_key = tuple(neighbors_of_interest)

            if neighbors_key in self._noise_matrices_dictionary[cluster_key]:
                deps_matrix = self._noise_matrices_dictionary[cluster_key][neighbors_key]
            else:
                deps_matrix = self._compute_noise_matrix_dependent(qubits_of_interest,
                                                                   neighbors_of_interest)

            for matrix_test in deps_matrix.values():
                if not qrem_math.is_matrix_stochastic(matrix_test):
                    raise ValueError(f"Noise matrix not stochastic for qubits: "
                                     f"cluster = {cluster_key}, neighbors = {neighbors_of_interest}.")

            return deps_matrix

    def compute_subset_noise_matrices_averaged(self,
                                               subsets_list: List[Tuple[int]],
                                               show_progress_bar: Optional[bool] = False) -> None:
        """Description:
        computes averaged (over all other qubits) noise matrices on subsets_list of qubits

        :param subsets_list: subsets_list of qubit indices
        :param show_progress_bar: whether to show animated progress bar. requires tqdm package

        """

        for subset_index in tqdm(subsets_list, disable = not show_progress_bar):
            
            #JT There is a bug in the line below, at least when it comes to data structures 
            #that we use as for now. For now I comment that line and replace it with a one that
            #works 
            #self._compute_noise_matrix_averaged(subsets_list[subset_index])

            self._compute_noise_matrix_averaged(subset_index)





# -*- coding: utf-8 -*-
"""
Created on Fri May 22 18:05:24 2015

@author: Peter Wittek
"""
from __future__ import division, print_function
import numpy as np
from sympy import expand
import time
from .nc_utils import pick_monomials_up_to_degree, simplify_polynomial, \
                      apply_substitutions, separate_scalar_factor, \
                      is_number_type
from .cvxpy_utils import solve_with_cvxpy
from .sdpa_utils import solve_with_sdpa, convert_row_to_sdpa_index, detect_sdpa
from .mosek_utils import solve_with_mosek
from .picos_utils import solve_with_cvxopt


def autodetect_solvers(solverparameters):
    solvers = []
    if detect_sdpa(solverparameters) is not None:
        solvers.append("sdpa")
    try:
        import cvxpy
    except ImportError:
        pass
    else:
        solvers.append("cvxpy")
        solvers.append("scs")
    try:
        import mosek
    except ImportError:
        pass
    else:
        solvers.append("mosek")
    try:
        import picos
    except ImportError:
        pass
    else:
        solvers.append("cvxopt")
    return solvers


def solve_sdp(sdp, solver=None, solverparameters=None):
    """Call a solver on the SDP relaxation. Upon successful solution, it
    returns the primal and dual objective values along with the solution
    matrices.

    :param sdpRelaxation: The SDP relaxation to be solved.
    :type sdpRelaxation: :class:`ncpol2sdpa.SdpRelaxation`.
    :param solver: The solver to be called, either `None`, "sdpa", "mosek",
                   "cvxpy", "scs", or "cvxopt". The default is `None`,
                   which triggers autodetect.
    :type solver: str.
    :param solverparameters: Parameters to be passed to the solver. Actual
                             options depend on the solver:

                             SDPA:

                               - `"executable"`:
                                 Specify the executable for SDPA. E.g.,
                                 `"executable":"/usr/local/bin/sdpa"`, or
                                 `"executable":"sdpa_gmp"`
                               - `"paramsfile"`: Specify the parameter file

                             Mosek:
                             Refer to the Mosek documentation. All
                             arguments are passed on.

                             Cvxopt:
                             Refer to the PICOS documentation. All
                             arguments are passed on.

                             Cvxpy:
                             Refer to the Cvxpy documentation. All
                             arguments are passed on.

                             SCS:
                             Refer to the Cvxpy documentation. All
                             arguments are passed on.
    :type solverparameters: dict of str.
    :returns: tuple of the primal and dual optimum, and the solutions for the
              primal and dual.
    :rtype: (float, float, list of `numpy.array`, list of `numpy.array`)
    """
    solvers = autodetect_solvers(solverparameters)
    solver = solver.lower() if solver is not None else solver
    if solvers == []:
        raise Exception("Could not find any SDP solver. Please install SDPA," +
                        " Mosek, Cvxpy, or Picos with Cvxopt")
    elif solver is not None and solver not in solvers:
        print("Available solvers: " + str(solvers))
        if solver == "cvxopt":
            try:
                import cvxopt
            except ImportError:
                pass
            else:
                raise Exception("Cvxopt is detected, but Picos is not. "
                                "Please install Picos to use Cvxopt")
        raise Exception("Could not detect requested " + solver)
    elif solver is None:
        solver = solvers[0]
    primal, dual, x_mat, y_mat, status = None, None, None, None, None
    tstart = time.time()
    if solver == "sdpa":
        primal, dual, x_mat, y_mat, status = \
          solve_with_sdpa(sdp, solverparameters)
    elif solver == "cvxpy":
        primal, dual, x_mat, y_mat, status = \
          solve_with_cvxpy(sdp, solverparameters)
    elif solver == "scs":
        if solverparameters is None:
            solverparameters_ = {"solver": "SCS"}
        else:
            solverparameters_ = solverparameters.copy()
            solverparameters_["solver"] = "SCS"
        primal, dual, x_mat, y_mat, status = \
          solve_with_cvxpy(sdp, solverparameters_)
    elif solver == "mosek":
        primal, dual, x_mat, y_mat, status = \
          solve_with_mosek(sdp, solverparameters)
    elif solver == "cvxopt":
        primal, dual, x_mat, y_mat, status = \
          solve_with_cvxopt(sdp, solverparameters)
        # We have to compensate for the equality constraints
        for constraint in sdp.constraints[sdp._n_inequalities:]:
            idx = sdp._constraint_to_block_index[constraint]
            sdp._constraint_to_block_index[constraint] = (idx[0],)
    else:
        raise Exception("Unkown solver: " + solver)
    sdp.solution_time = time.time() - tstart
    sdp.primal = primal
    sdp.dual = dual
    sdp.x_mat = x_mat
    sdp.y_mat = y_mat
    sdp.status = status
    return primal, dual, x_mat, y_mat


def find_solution_ranks(sdp, xmat=None, baselevel=0):
    """Helper function to detect rank loop in the solution matrix.

    :param sdp: The SDP relaxation.
    :type sdp: :class:`ncpol2sdpa.sdp`.
    :param x_mat: Optional parameter providing the primal solution of the
                  moment matrix. If not provided, the solution is extracted
                  from the sdp object.
    :type x_mat: :class:`numpy.array`.
    :param base_level: Optional parameter for specifying the lower level
                       relaxation for which the rank loop should be tested
                       against.
    :type base_level: int.
    :returns: list of int -- the ranks of the solution matrix with in the
              order of increasing degree.
    """
    if sdp.status == "unsolved" and xmat is None:
        raise Exception("The SDP relaxation is unsolved and no primal " +
                        "solution is provided!")
    elif sdp.status != "unsolved" and xmat is None:
        xmat = sdp.x_mat[0]
    else:
        xmat = sdp.x_mat[0]
    if sdp.status == "unsolved":
        raise Exception("The SDP relaxation is unsolved!")
    ranks = []
    from numpy.linalg import matrix_rank
    if baselevel == 0:
        levels = range(1, sdp.level + 1)
    else:
        levels = [baselevel]
    for level in levels:
        base_monomials = \
            pick_monomials_up_to_degree(sdp.monomial_sets[0], level)
        ranks.append(matrix_rank(xmat[:len(base_monomials),
                                      :len(base_monomials)]))
    if xmat.shape != (len(base_monomials), len(base_monomials)):
        ranks.append(matrix_rank(xmat))
    return ranks


def get_sos_decomposition(sdp, y_mat=None, threshold=0.0):
    """Given a solution of the dual problem, it returns the SOS
    decomposition.

    :param sdp: The SDP relaxation to be solved.
    :type sdp: :class:`ncpol2sdpa.sdp`.
    :param y_mat: Optional parameter providing the dual solution of the
                  moment matrix. If not provided, the solution is extracted
                  from the sdp object.
    :type y_mat: :class:`numpy.array`.
    :param threshold: Optional parameter for specifying the threshold value
                      below which the eigenvalues and entries of the
                      eigenvectors are disregarded.
    :type threshold: float.
    :returns: The SOS decomposition of [sigma_0, sigma_1, ..., sigma_m]
    :rtype: list of :class:`sympy.core.exp.Expr`.
    """
    if len(sdp.monomial_sets) != 1:
        raise Exception("Cannot automatically match primal and dual " +
                        "variables.")
    elif len(sdp.y_mat[1:]) != len(sdp.constraints):
        raise Exception("Cannot automatically match constraints with blocks " +
                        "in the dual solution.")
    elif sdp.status == "unsolved" and y_mat is None:
        raise Exception("The SDP relaxation is unsolved and dual solution " +
                        "is not provided!")
    elif sdp.status != "unsolved" and y_mat is None:
        y_mat = sdp.y_mat
    sos = []
    for y_mat_block in y_mat:
        term = 0
        vals, vecs = np.linalg.eigh(y_mat_block)
        for j, val in enumerate(vals):
            if val < -0.001:
                raise Exception("Large negative eigenvalue: " + val +
                                ". Matrix cannot be positive.")
            elif val > 0:
                sub_term = 0
                for i, entry in enumerate(vecs[:, j]):
                    sub_term += entry * sdp.monomial_sets[0][i]
                term += val * sub_term**2
        term = expand(term)
        new_term = 0
        if term.is_Mul:
            elements = [term]
        else:
            elements = term.as_coeff_mul()[1][0].as_coeff_add()[1]
        for element in elements:
            _, coeff = separate_scalar_factor(element)
            if abs(coeff) > threshold:
                new_term += element
        sos.append(new_term)
    return sos


def get_facvar_of_monomial(monomial, sdp):
    results = sdp._get_index_of_monomial(monomial)
    if sdp._new_basis is None:
        facvar = np.zeros((sdp.n_vars + 1,))
    else:
        facvar = np.zeros((len(sdp._original_obj_facvar) + 1,))
    for (k, coeff) in results:
        facvar[k] += coeff
    if sdp._new_basis is not None:
        new_facvar = np.zeros((sdp.n_vars + 1,))
        new_facvar[0] = facvar[0]
        new_facvar[1:] = sdp._new_basis.T.dot(facvar[1:])
        facvar = new_facvar
    return facvar


def get_recursive_xmat_value(k, row_offsets, sdp, x_mat):
    Fk = sdp.F[:, k]
    for row in range(len(Fk.rows)):
        if Fk.rows[row] != []:
            block, i, j = convert_row_to_sdpa_index(sdp.block_struct,
                                                    row_offsets, row)
            value = x_mat[block][i, j]
            for index in sdp.F.rows[row]:
                if k != index:
                    value -= sdp.F[row, index] * \
                        get_recursive_xmat_value(index, row_offsets,
                                                 sdp, x_mat)
            return value / sdp.F[row, k]


def get_xmat_value(monomial, sdp, x_mat=None):
    if sdp.status == "unsolved" and x_mat is None:
        raise Exception("The SDP relaxation is unsolved and no primal " +
                        "solution is provided!")
    elif sdp.status != "unsolved" and x_mat is None:
        x_mat = sdp.x_mat
    polynomial = expand(simplify_polynomial(monomial,
                                            sdp.substitutions))
    if polynomial.is_Mul:
        elements = [polynomial]
    else:
        elements = polynomial.as_coeff_mul()[1][0].as_coeff_add()[1]
    row_offsets = [0]
    cumulative_sum = 0
    for block_size in sdp.block_struct:
        cumulative_sum += block_size ** 2
        row_offsets.append(cumulative_sum)
    linear_combination = get_facvar_of_monomial(0, sdp)
    for element in elements:
        element, coeff = separate_scalar_factor(element)
        if element in sdp.moment_substitutions:
            element = sdp.moment_substitutions[element]
        element = apply_substitutions(element, sdp.substitutions)
        linear_combination += coeff*get_facvar_of_monomial(element, sdp)
    value = linear_combination[0]
    linear_combination[0] = 0
    # This used to be a conditional when moments were never substituted:
    # if len(sdp.block_struct) > sdp.constraint_starting_block:
    for row in range(row_offsets[sdp.constraint_starting_block]):
        intersect = np.intersect1d(sdp.F.rows[row],
                                   np.nonzero(linear_combination)[0])
        if len(intersect) == 0:
            continue
        elif len(intersect) == 1:
            block, i, j = convert_row_to_sdpa_index(sdp.block_struct,
                                                    row_offsets, row)
            value += linear_combination[intersect[0]]*x_mat[block][i, j]
            linear_combination[intersect[0]] = 0
            for index in sdp.F.rows[row]:
                if index not in intersect:
                    value -= sdp.F[row, index] * \
                        get_recursive_xmat_value(index, row_offsets,
                                                 sdp, x_mat)
        else:
            # This can only happen because we changed the basis
            A = np.zeros((len(intersect), len(intersect)))
            b = np.zeros((len(intersect), ))
            in_ = 0
            rank0 = 0
            for row2 in range(row,
                              row_offsets[sdp.constraint_starting_block]):
                block, i, j = convert_row_to_sdpa_index(sdp.block_struct,
                                                        row_offsets, row2)
                is2 = np.intersect1d(sdp.F.rows[row2],
                                     np.nonzero(linear_combination)[0])
                if len(is2) == len(intersect):
                    col = 0
                    for it in intersect:
                        A[in_, col] = sdp.F[row2, it]
                        col += 1
                    rank1 = np.linalg.matrix_rank(A)
                    b[in_] = x_mat[block][i, j]
                    if rank1 > rank0:
                        rank0 = rank1
                        if in_ == len(intersect) - 1:
                            break
                        in_ += 1
            x = np.linalg.solve(A, b)
            for k, it in enumerate(intersect):
                value += x[k]*linear_combination[it]
                linear_combination[it] = 0
    return value


def extract_dual_value(sdp, monomial, blocks=None):
    """Given a solution of the dual problem and a monomial, it returns the
    inner product of the corresponding coefficient matrix and the dual
    solution. It can be restricted to certain blocks.

    :param sdp: The SDP relaxation.
    :type sdp: :class:`ncpol2sdpa.sdp`.
    :param monomial: The monomial for which the value is requested.
    :type monomial: :class:`sympy.core.exp.Expr`.
    :param monomial: The monomial for which the value is requested.
    :type monomial: :class:`sympy.core.exp.Expr`.
    :param blocks: Optional parameter to specify the blocks to be included.
    :type blocks: list of `int`.
    :returns: The value of the monomial in the solved relaxation.
    :rtype: float.
    """
    if sdp.status == "unsolved":
        raise Exception("The SDP relaxation is unsolved!")
    if blocks is None:
        blocks = [i for i, _ in enumerate(sdp.block_struct)]
    if is_number_type(monomial):
        index = 0
    else:
        index = sdp.monomial_index[monomial]
    row_offsets = [0]
    cumulative_sum = 0
    for block_size in sdp.block_struct:
        cumulative_sum += block_size ** 2
        row_offsets.append(cumulative_sum)
    result = 0
    for row in range(len(sdp.F.rows)):
        if len(sdp.F.rows[row]) > 0:
            col_index = 0
            for k in sdp.F.rows[row]:
                if k != index:
                    continue
                value = sdp.F.data[row][col_index]
                col_index += 1
                block_index, i, j = convert_row_to_sdpa_index(
                    sdp.block_struct, row_offsets, row)
                if block_index in blocks:
                    result += -value*sdp.y_mat[block_index][i][j]
    return result



"""Constants used in visualisation module, e.g. colors, line widths etc."""

    

try:
    from manim import config, WHITE, GRAY_B, GRAY_C
    # manim configuration:
    config.background_color = WHITE
except:
    WHITE = '#FFFFFF'
    GRAY_B = 	'#777777'
    GRAY_C = '#444444'
     
import matplotlib as mpl
from matplotlib.colors import LinearSegmentedColormap, to_hex



# graph constants:
radius_size = 0.36
arrow_tip_size = 0.5
buff_val = 0.30
stroke_width_thin = 4
stroke_width_medium = 5
stroke_width_thick = 6
max_stroke_width_to_length_ratio = 50

# legend constants:
legend_bar_height = 6
legend_bar_width = 0.3

# colors:
opacity_of_unused = 1
lightest_gray = '#F5F5F5'
medium_gray = GRAY_B
darkest_gray = GRAY_C
cmap = mpl.cm.viridis
colors = [to_hex(cmap(1.0)), to_hex(cmap(0.75)), to_hex(cmap(0.5)), to_hex(cmap(0.5)), to_hex(cmap(0.25)),
          to_hex(cmap(0.25)), to_hex(cmap(0.25)), to_hex(cmap(0.)), to_hex(cmap(0.)), to_hex(cmap(0.))]
color_map = LinearSegmentedColormap.from_list("color_map_viridis", colors)
